///////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool - A program for Multivariate Data Display
//
//  Copyright 1994-2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
///////////////////////////////////////////////////////////// COPYRIGHT //

//////////////////////////////////////////////////////////////////////////
/// \file ResultSet.h
/// \brief Defines the interface for the resultset.
//////////////////////////////////////////////////////////////////////////

#ifndef _RESULTSET_H
#define _RESULTSET_H

#include<string>
#include<iostream>

//#define OTL_ODBC_UNIX

#include"odbc.h"

//Forward References
class DataValue;
class otl_stream;

using namespace std;

//////////////////////////////////////////////////////////////////////////
/// \brief 
///
//////////////////////////////////////////////////////////////////////////
class ResultSet
{
public:		// methods

	/// \name Structors
	//@{
	/// Constructor
	 ResultSet(otl_stream *otlStream);

	/// Destructor
	~ResultSet();
	//@}

	/** Returns the current row number.
	* @return The current row number in the result set, or <code>0</code>
	* if it can't be determined.
	*/
	int getRow();

	/**
	* Returns the number of columns in the resultSet.
	* \return numOfColumns
	*/
	int getNumberOfColumns();

	/**
	* Returns true if the result set is empty.
	* \return bool
	*/
	bool isDone();


	/** Moves to the next row in the result set 
	* @return true if the cursor is in the result set
	*/
	bool next();


	/** Gets a column's value as a double
	* @param idx The column index, starting at 1
	*/
	double getDouble(int idx);


	/** Gets a column's value as a float
	* @param idx The column index, starting at 1
	*/
	float getFloat(int idx);

	/** Gets a column's value as a Long
	* @param idx The column index, starting at 1
	*/
	long getLong(int idx);
	
	/** Gets a column's value as an int
	* @param idx The column index, starting at 1
	*/
	int getInt(int idx);


	/** Gets a column's value as a string
	* @param idx The column index, starting at 1
	*/
	std::string getString(int idx);	

	/*
	 * This function returns a pointer of the otl_column_desc 
	* type to the descriptor of the output column list: 
	* The pointer points to an internal structure inside the stream which 
	* gets deallocated at the moment of the stream desctruction, so the user does not need to do any memory deallocation operations with the pointer. 
	* @The desc_len returns the lenght of the allocated array.
	*/
	otl_column_desc* describe_select(int& desc_len);

	private://methods

	/**
	* This populates the row with a collection of data values.
	* It uses the columnDesc returned from the data base
	* @post The calling procedure needs to remove call delete on 
	* each of the data values
	* \param inputStream 
	* \param *row 
	*\return int, Returns the number of datavalues in the row
	*/
	int makeDataValue(otl_stream &inputStream,DataValue ***row);
	
	private:	// members
		otl_stream *otlStream;
		DataValue **row;
		int numOfColumns;

};


class DataValue
{
	/**
	* Creates a Data Value for a particular type
	* \param t 
	* \return 
	*/
public: 
	enum Type {integer=0,lInteger=1,real=3,lReal=4,str=5};

	DataValue(Type t,int size)
	{
		this->t = t;
		this->size = size;
		switch(t)
		{
		case integer:
			this->data_value = (void *) new int;
			break;
		case lInteger:
			this->data_value = (void *) new long;
			break;
		case real: 
			this->data_value = (void *) new float;
			break;
		case lReal:
			this->data_value = (void *) new double;
			break;		
		case str:
			this->data_value = (void *) new std::string;
			break;
		} // switch(t)

	}
	/**
	*
	*/
public :
	/**
	* Returns the data value as a void *
	* @pre, The set_value method should be called before this is called.
	* @post, This class is not responsible for maintaining the value
	* at the pointer. Therefore if required it should be copied to another variable
	*/
	void *getValue()
	{
		return(this->data_value);
	}

	/**
	*
	* sets the value of the variable
	*@pre the inputstream should not be empty
	* \return 
	*/
	void setValue(otl_stream & inputStream)
	{
		try{
		switch(t)
		{
			//Integer
		case integer:
		{
			int tmp_int = *((int*)this->data_value);
			inputStream.operator>>(tmp_int);
			inputStream >> tmp_int;
			break;
		}
		case lInteger:
		{
			long tmp_long = *((long*)this->data_value);
			inputStream>> tmp_long;
			break;
		}
		case real:
		{
			float tmp_float = *((float*)this->data_value);
			inputStream>> tmp_float;
			break;
		}
		case lReal:
		{
			double tmp_double = *((double*)this->data_value);
			inputStream>> tmp_double;
			break;		
		}
		case str:
		{
			char *temp = new char[this->size + 1];
			inputStream>>temp;
			std::string tempString(temp);
			(*(std::string*)this->data_value) = ""+tempString;			
			delete[] temp;
			break;
		}
		}} catch(otl_exception& p)
		{
			std::cerr<<p.code<<std::endl; // print out error code
			std::cerr<<p.sqlstate<<std::endl; // print out error SQLSTATE
			std::cerr<<p.msg<<std::endl; // print out error message
			std::cerr<<p.stm_text<<std::endl; // print out SQL that caused the error
			std::cerr<<p.var_info<<std::endl; // print out thevariable that caused the error
		}
	}

	/**
	* Destroys the data value.
	* That is deletes the object allocated.
	* \return 
	*/

	~DataValue()
	{
		switch(t)
		{
			//Integer
		case integer:
			delete((int*)this->data_value);
			break;
		case lInteger:
			delete((long*)this->data_value);
			break;
		case real:
			delete((float*)this->data_value);
			break;
		case lReal:
			delete((double*)this->data_value);
			break;
		case str:
			delete((std::string*)this->data_value);
		}

	}

	//Stores the private variables
private :
	enum Type t;
	void *data_value;
	int size;
	
};

#endif	// _RESULTSET_H

