#ifndef _CACHE_CONTROLLER_H
#define _CACHE_CONTROLLER_H

#include <ILoadListener.h>
#include <IReplacementVisitor.h>
#include <ICacheController.h>
#include <log/log.h>
#include <cache/PerformanceStatistics.h>
#include <cache/IFragThresholdVisitor.h>

namespace cache
{

	//Forward References
	class CacheManager;
	class BrushStore;
	class PrefetchController;
	class ICacheEntry;

	template <class Item>
	class Iterator;

	/*
	*	CacheController provides an interface to the Cache
	*  controls the actions of the prefetcher.
	*  Comments: if(pc!= NULL) is not an elegant solution a better design would
	*  be to have the 2 versions of the class one with prefetcher and the other
	*  without prefetcher, then have a common pure virutal class on top of it.
	*  But I am choosing this way for efficiency purposes
	*/
	class CacheController : public IReplacementVisitor,
		public ICacheController, public IFragThresholdVisitor
	{
	public:
		/**
		* Constrcuts a new cache controller
		* \param PrefetchController 
		* \param *pc 
		* \return 
		*/
		CacheController(CacheManager* cacheManager,
			const unsigned int numberOfDimensions,					
			BrushStore* brushStore,
			PerformanceStatistics* perStatistics,
			PrefetchController *pc=NULL);


		

		/**
		* Adds the listener to listen to the start and stop load events
		* \param loadListener 
		*/
		void addLoadListener(ILoadListener* loadListener);



		/**
		* Request from the Front End
		* \param e1 
		* \param e2 
		* \param lod 
		* \param type 
		* \return 
		*/
		::Iterator<ICacheEntry*>* load(double e1,double e2,double lod,int type);
		
		/*
		* This returns the performance statistics for
		* the last time the load method was called
		* in msec
		*	@return unsigned long
		*/
		PerformanceStatistics* getPerformanceStatistics();
		/**
		* Called by the DataIterator when it is being deleted
		*/
		void endLoad();

		
		/*
		 * Returns the current prefetch controller used only for experimental purposes
		 *	@return PrefetchController*
		 */
		PrefetchController* CacheController::getPrefetchController();

		//implements the IReplacementVisitor interface
		/**
		* 
		* \param cacheEntry 
		* \return 
		*/
		bool chosenItem(ICacheEntry& cacheEntry);
		
		/*
		 *	Implements the IFragThresholdVisitor Interface
		 *  This is used to notify the cache controller that the number of
		 *  fragments for the query has exceeded the limit.
		 * 
		 */
		bool exceededFragThreshold();

		
		~CacheController();

	private:
		//Stores the reference to the prefetch controller.
		PrefetchController* pc;
		//Stores the list of load listeners
		std::vector<ILoadListener*> vecLoadListener;
		//Stores the pointer to the CacheManager
		CacheManager* cacheManager;
		//Stores the reference to the brush store
		BrushStore* brushStore;
		//Stores the number of dimensions
		unsigned long numberOfDimensions;
		//Stores the performance statistics update after each call to load
		PerformanceStatistics* perStatistics;
		//Logger used to log the messages
		debug::log::CMessageGroup controllerLogger;
	};
}
#endif //_CACHE_CONTROLLER_H