#ifndef _PREFETCHER_H
#define _PREFETCHER_H

#include <cache/CacheManager.h>
#include <zthread/Thread.h>
#include <zthread/Semaphore.h>
#include <db_GLOBAL.h> //Contains the definitions for all the macros
#include <strategyTable.h> //Defines the methods for the strategy table
#include <db_DISCOVER1.h>
#include <cache/PrefetcherHelper.h>
#include <log/log.h>



namespace cache
{
	//Forward References
	class Brush;
	class PrefetchController;	
	class PrefetchVisitor;
	class LoadRequester;
	class BrushStore;
	class ICacheEntry;
	class AbstractRegionFactory;
	class CacheIterator;
	class QueryModifier;
	

	class Prefetcher : public ZThread::Runnable 
	{
		friend class PrefetchController;
	public: 

		Prefetcher(const unsigned long sizeOfCache,
			unsigned int numberOfLevels,//Used temporarily
			BrushStore* brushStore,		
			int AdaptivePrefetching = 1,
			PrefetchVisitor *prefetchVisitor = NULL,
			unsigned int PrefetchStrategy = 1, //Initial Prefetch Strategy 1 for random
			unsigned int NO_P_ENTRIES = 11//This is also temporary		
			);
		
		/**
		* Destroys the Prefetcher
		* \return 
		*/
		virtual ~Prefetcher();

		/**
		* Starts the Prefetcher
		*/
		void run();
		
		/**
		* Used by the prefetch Controller to preempt the prefetcher
		* \param preEmpt 
		*/
		void setPreEmpt(bool preEmpt);
		
		/*
		*	Returns the pointer to the prefetch brushes.
		*/
		::std::vector<Brush*>& getPrefetchBrushes();
		
		/**
		*
		* \Decides whether the prefetcher wants to continue or not 
		* Note it depends upon the current replacement candidate whether
		* the prefetcher want to continue or not.
		* \return bool
		*/	
		bool ifContinue(ICacheEntry& cacheEntry);
		

		/**
		* Returns the number of nodes prefetched by the prefetcher
		* \return unsigned int
		*/
		unsigned int getNumberOfNodesPrefetched();
		
	protected:

		//Methods to maintain a consistent state
		void cleanup();
		

		//Methods imported from old prefetcher
		/*************************************************************************
		init_dir

		If prefetch strategy is random, make discover paths in all
		the directions=0.25.
		For the remaining strategies make it 0.1
		*************************************************************************/

		void init_dir(double *vect);
		


		/*************************************************************************
		adjust_prefetch_values

		Depending on the direction in which to prefetch,
		assign the focus values from the focus table.
		*************************************************************************/

		void adjust_prefetch_values(int direction, double *up, double *dn,
									double *le, double *ri);
		
		/*************************************************************************
		calculateStatisticsForFitnessFunction

		This function counts the number of objects predicted by the
		given prefetching scheme

		*************************************************************************/
		/*
		*	This function will change completely because we do 
		* have a different cacheStrcuture and we cannot do something like this
		* This same task is acheived in the following manner
		* a) Summing over the number of hits each time we call the load in the prefetch
		* b) Summing over the number of new items loaded this can be done in ifContinue()
		*/
		void calculateStatisticsForFitnessFunction(double x, double y, double level);
		

		/*************************************************************************
		do_prefetching

		Call init_dir()
		For prefetch strategy=2 or 3, set discover paths to 0.6
		if there are more regions to be discovered in that direction.

		start loop

		calculate the probability in the direction in which to move.
		if prefetch strategy is 'focus', then adjust the prefetch
		values by calling adjust_prefetch_values()

		determine what data to be loaded from the database into memory
		in each directions by calling preparePrefetching() 
		separately for each cirections.

		end loop

		(ger) revised to use preparePrefetching instead of prepare() to work with
		or without incremental loading
		*************************************************************************/

		void do_prefetching();

		/*
		 *	@return, Returns the number of nodes prefetched.
		 */
		void loadBrush(double leftExtent,
			double rightExtent,
			double lod,
			AbstractRegionFactory* regionFactory,
			QueryModifier& queryModifier);

		//Sets the semaphore
	private:
		//Set by the Prefetch Controller
		void setSemaphore(ZThread::Semaphore* semPrefetchPermit,
			ZThread::Semaphore* semLoadPermit);
		
	private: 	
		//Stores the size of the cache
		unsigned long sizeOfCache;	
		//Stores the flag
		bool flag;
		//Just used temporarily
		unsigned int numberOfLevels;
		//Stores the number of hot regions
		int DISC_FOC;	/* Number of hot regions */

		//Variables imported from old prefetcher

		/* Legend:
		*      index -> hot region #
		*      xmdvFocus[index][0] -> left  extent
		*      xmdvFocus[index][1] -> right extent
		*      xmdvFocus[index][2] -> level extent
		*/	

		//Stores the initial prefetch strategy, used to be set by appsetPrefetchStrategy
		unsigned int Prefetch_Strategy;
		//stores the foccus area
		double **xmdvFocus;	
		//Also we need to the current prefetch brush
		BrushStore* brushStore;
		//Stores the number of P Entries
		int NO_P_ENTRIES;
		//Stores the reference to the abstract factory
		AbstractRegionFactory *regionFactory;
		//Stores the type of the prefetch brush
		int brushType;
		//Stores the current cache iterator
		CacheIterator* cacheIterator;
		//Stores the vector of prefetch brushes, in most cases it will contain one element
		::std::vector<Brush*> vecPrefetchBrushes;
		//Stores the PrefetcherHelper
		PrefetcherHelper prefetchHelper; 
		//Store the flag
		bool preempt;
		//Stores the total number of nodes fetched in this pass
		unsigned int numberOfNodesPrefetched;
		//Used for debugging purposes
		debug::log::CMessageGroup prefetchLogger;	
		//Stores the reference to the prefetch Visitor
		PrefetchVisitor*prefetchVisitor;
		//Store if Adaptive Prefetching is enabled or not
		int DO_ADAPTIVE_PREFETCHING;
		//Store the semaphore for thread synchronization
		ZThread::Semaphore *semPrefetchPermit;
		//Stores the semaphore for the load Permit
		ZThread::Semaphore *semLoadPermit;

	};
}
#endif //PREFETCHER_H