////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//  	Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//  	Computer Science Department
//  	Worcester Polytechnic Institute
//  	Worcester, MA 01609 USA
//
//  	http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////// COPYRIGHT //

#ifndef COLORMAP_H
#define COLORMAP_H

// SYSTEM INCLUDES
//
#include <string>

// PROJECT INCLUDES
//
#include <xmdv/RGB.h>
#include <xmdv/colorassign.h>
#include <xmdv/Visitor.h>

// CONSTANTS
//
const int CMAPSIZE	= 256;

// FORWARD REFERENCES
//
class ColorRamp;
class Image;

///
/// Color map or Color LUT (look-up table). Specified as per the color
/// ramp being used.
///
/// USAGE: 
/// <pre>
/// 	colormap map;
///		map.SetColorRamp(colorramp::GetPredefined2());
///		map->DrawColorBar(0,0,10,0,5);
///		unsigned long color = map[50];
/// </pre>
/// 
/// \date 02-05-2004
/// \author anilkpatro
/// \version 1.0
///
class colormap
{
public:
	/// Color models currently supported by the color map
	enum ColorModel { 
		MODEL_RGB, 
		MODEL_GREY,
		MODEL_HSV, 
	};

public:		// methods

// LIFECYCLE

	/// 
	/// Constructor - initialize the color map with the specified ramp
	/// and color model
	/// 
	/// \parma name name of the colormap
	/// \param ramp curve
	/// \param len length of the map
	/// \param model color model
	///
	colormap(const std::string &name, ColorRamp *ramp = 0, 
			 int len = CMAPSIZE, ColorModel model = MODEL_RGB);

	/// 
	/// Copy Constructor
	/// 
	/// \param from copy from this object
	/// \return 
	///
	colormap(const colormap &from);
	
	/// 
	/// Destructor
	///
	virtual		~colormap();

// OPERATORS

	/// 
	/// Assignment operator
	/// 
	/// \param from copy from this object
	///
	void		operator =(const colormap & from);

	/// 
	/// Indexed access to the colors in the color map
	/// 
	/// \param index index of the color [0, Len())
	/// \return rgb color as long
	///
	unsigned long	operator[](int index) const;

	/// 
	/// Indexed access to the RGB color in the color map
	/// 
	/// \param index index of the color [0, Len())
	/// \return RGB color
	///
	RGBt		GetRGBAt(int index) const;

	/// 
	/// Indexed access to the HSI color in the color map
	/// 
	/// \param index index of the color [0, Len())
	/// \return HSI color 
	///
	HSIColor	GetHSIAt(int index) const;

// OPERATIONS

	void		Accept(Visitor<colormap>& visitor);

	/// 
	/// Draw the color map
	/// 
	/// \param x1 x position of the initial point
	/// \param y1 y position of the initial point
	/// \param x2 x position of the final point
	/// \param y2 y position of the final point
	/// \param width width of the drawn color map
	///
	void		DrawColorBar(float x1 = 0.0f, float y1 = 0.0f,
							 float x2 = 1.0f, float y2 = 0.0f,
							 float width = 5.0f);

	/// 
	/// Initialize the texture for drawing
	///
	void		InitTex();

	/// 
	/// Load a colormap from the inputstream
	/// 
	/// \param inputstream stream to load the colormap from
	/// \param cmapname name of the new colormap
	/// \return the loaded colormap
	///
	static colormap *Load(std::istream &inputstream, 
						  const std::string &cmapname);
	
	/// 
	/// Save the specified colormap to the stream
	/// 
	/// \param cmap the colormap
	/// \param outputstream the stream to save to
	/// \return status (true = saved successfully)
	///
	static bool		Save(colormap *cmap, std::ostream &outputstream);

// ACCESS

	/// 
	/// Get the name of this colormap
	/// 
	/// \return name
	///
	std::string	GetName() const;
	/// 
	/// Set the name of this colormap
	/// 
	/// \param name name of the colormap
	///
	void		SetName(const std::string &name);

	/// 
	/// Get the length of the color map
	/// 
	/// \return length of the color map
	///
	int			Len() const;

	/// 
	/// Get the ramp that's being currently used for the map
	/// 
	/// \return the ramp
	///
	ColorRamp	*GetColorRamp();
	/// 
	/// Set the ramp(curve) to be used for the map
	/// 
	/// \param ramp color ramp to be used for the map
	///
	void		SetColorRamp(ColorRamp *ramp);

	ColorModel	GetColorModel() const;
	void		SetColorModel(ColorModel model);

private:	// members

	std::string	m_name;			///< name of the color map

	ColorRamp	*m_colorramp;	///< the color ramp
	int			m_len;			///< length of the color map
	ColorModel	m_model;		///< color model for the map

	Image		*m_checkImg;	///< checkered texture for drawing

};

//-------------------------------------------------------------------------
inline int colormap::Len() const
{
	return m_len;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
inline ColorRamp * colormap::GetColorRamp()
{
	return m_colorramp;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
inline std::string colormap::GetName() const
{
	return m_name;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
inline colormap::ColorModel colormap::GetColorModel() const
{
	return m_model;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
inline void colormap::Accept(Visitor<colormap>& visitor)
{
	visitor.Visit(*this);
}
//-------------------------------------------------------------------------

#endif
