#include "som.h"



SOM::SOM(int N, int M): 
	topol(0),
	neigh(0),
	alpha_type(0),
	mapdist(NULL),
	neigh_adapt(NULL),
	alpha_func(NULL),
	radius(0.0),
	alpha(0.0),
	length(0),
	knn(0),
	codes(N, M)
{}

void SOM::randinit(const SOM_Matrix& trdata, int topoltype, int nei, int xdim, int ydim)
{
	data=trdata;
	topol=topoltype;
	neigh=nei;
	codes.setSize(xdim, ydim);


	SOM_Vector minvalue(data.len1), maxvalue(trdata.len1);
	minvalue=trdata.minrow();
	maxvalue=trdata.maxrow();
	for (int n=0; n<codes.len0; n++) {
		for (int m=0; m<codes.len1; m++) {
			codes(n, m).setSize(trdata.len1);
			codes(n, m).random(minvalue, maxvalue);
		}
	}	
}


int SOM::setCodes(const SOM_Matrix& codesdata)
{
	if (codesdata.len0!=codes.len0*codes.len1) return 0;

	for (int n=0; n<codes.len0; n++)
		for (int m=0; m<codes.len1; m++)
			codes(n, m)=codesdata.row(n*codes.len1+m);

	return 1;
}



void SOM::setparam(short alp_type, double rad, double alp, long len)
{
	alpha_type=alp_type;
	rad=radius;
	alpha=alp;
	length=len;

	// set function pointers
	/*
	mapdist=dist;

	if (neigh==0) neigh_adapt = bubble_adapt;
	else neigh_adapt = gaussian_adapt;

	if (alpha_type==0) alpha_func = linear_alpha;
	else alpha_func = inverse_t_alpha; 
	*/
}


void SOM::training(short alpha_type, double radius, double alpha, long length)
{
	setparam(alpha_type, radius, alpha, length);

	int cnt=length/(data.len0);
	int mod=length%(data.len0);
	int count=0;

	for (int le=0; le<cnt; le++) {
		for (int i=0; i<data.len0; i++) {

			
			//radius decrease linearly to one
			double trad=1.0+(radius-1.0)*(length-count)/length;
			double talp;
			if (alpha_type==0) talp = linear_alpha(count, length, alpha);
			else talp = inverse_t_alpha(count, length, alpha);

			count++;
			//cout<<"length="<<length<<'\t'<<count<<endl;

			//search the best matching vectors -- minimum distance
			IP_MatrixIndex argmin=codes.minDist(data.row(i));

			if (neigh==0) bubble_adapt(data.row(i), argmin, trad, talp);
			else gaussian_adapt(data.row(i), argmin, trad, talp);
		}
	}
	for (int i=0; i<mod; i++) {

		
		//radius decrease linearly to one
		double trad=1.0+(radius-1.0)*(length-count)/length;
		double talp;
		if (alpha_type==0) talp = linear_alpha(count, length, alpha);
		else talp = inverse_t_alpha(count, length, alpha);

		count++;
		//cout<<"length="<<length<<'\t'<<count<<endl;

		//search the best matching vectors -- minimum distance
		IP_MatrixIndex argmin=codes.minDist(data.row(i));

		if (neigh==0) bubble_adapt(data.row(i), argmin, trad, talp);
		else gaussian_adapt(data.row(i), argmin, trad, talp);
	}
}


IP_Vector<IP_MatrixIndex> SOM::compute(const SOM_Matrix& dat)
{
	IP_Vector<IP_MatrixIndex> result(dat.len0);
	for (int i=0; i<dat.len0; i++) 
		result(i)=codes.minDist(dat.row(i));
	return result;
}

IP_Matrix<double> SOM::mapped_coord(const SOM_Matrix& dat)
{
	IP_Vector<IP_MatrixIndex> result(dat.len0);
	for (int i=0; i<dat.len0; i++) 
		result(i)=codes.minDist(dat.row(i));

	IP_Matrix<double> mapped(result.len, 2);
	for (i=0; i<result.len; i++) {
		mapped(i,0)= (double) result(i).row;
		mapped(i,1)= (double) result(i).col;
	}

	return mapped;
}

std::istream& operator >>(std::istream& is, SOM& s)
{
	return is;
}

std::ostream& operator <<(std::ostream& os, const SOM& s)
{
	char tbl='\t';
	for (int n=0; n<s.codes.len0; n++)
		for (int m=0; m<s.codes.len1; m++) 
			os<<n<<tbl<<m<<tbl<<s.codes(n,m)<<std::endl;
	return os;
}



double SOM::dist(int r, int c, int row, int  col)
{
	return (double) sqrt((double)((row-r)*(row-r)+(col-c)*(col-c)));
}



/* Adaptation function for bubble-neighborhood */

void SOM::bubble_adapt(const SOM_Vector& sample, const IP_MatrixIndex& winner, double radius, double alpha)
{
	for (int i=0; i<codes.len0; i++) 
		for (int j=0; j<codes.len1; j++) 
			if (dist(i, j, winner.row, winner.col) <= radius) 
				codes(i, j) +=alpha*(sample-codes(i,j));
}


/* Adaptation function for gaussian neighbourhood */
     
void SOM::gaussian_adapt(const SOM_Vector& sample, const IP_MatrixIndex& winner, double radius, double alpha)
{
	for (int i=0; i<codes.len0; i++) {
		for (int j=0; j<codes.len1; j++) {
			double dd = dist(i, j, winner.row, winner.col);
		    double alp = alpha * (double) exp((double) (-dd * dd / (2.0 * radius * radius)));
			codes(i, j) +=alp*(sample-codes(i,j));
		}
    }
}


/* linearly decreasing alpha */

double SOM::linear_alpha(long iter, long length, double alpha)
{
  return (alpha * (double) (length - iter) / (double) length);
}


/* inverse_t type of decreasing alpha */
#define INV_ALPHA_CONSTANT 100.0

double SOM::inverse_t_alpha(long iter, long length, double alpha)
{
  double c;
  
  c = length / INV_ALPHA_CONSTANT;
  
  return (alpha * c / (c + iter));
}



