// Display data using parallel coordinates.  Each dimensional axis is layed
// out in a vertical line, evenly spaced between dimensions.  The top and
// bottom of each line coincide with the maximum and minimum for that
// dimension.  Each data point creates a polyline across these axes.  Labels
// are printed along the top of each dimension, and brush coverage display
// consists of a single shaded region across the drawing area.

/*============================== INCLUDES ==============================*/

#include <xmdv/CFlatParcoord.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/brush_oper.h>
#include <xmdv/multi_select.h>
#include <xmdv/ogl.h>
#include <xmdv/brush.h>
#include <xmdv/CDimReducer.h>
#include <xmdv/CDimReductionDisplayInformation.h>

#ifdef USE_META
#include <meta/Meta.h>
#endif // USE_META

CFlatParcoord::CFlatParcoord(CFlatDisplayInformation *flat_info)
	: CFlatDisplay(flat_info)
{
	axis = -1;
}

CFlatParcoord::~CFlatParcoord()
{
}


// Draw the brush boundary in the parallel coordinate view.
void CFlatParcoord::DrawParcoordBrushBoundRamp(grCanvas *canvas,
											   Brush *brush,
											   unsigned long color1,
											   unsigned long color2)
{
	// set the appropriate color
	//
	ARR<Vec2> poly(g_globals.dataset->dims);

	double factor = 0.1;
	RGBt c1( color1), c2(color2);
	c2.R = factor * c1.R + (1 - factor) * c2.R;
	c2.G = factor * c1.G + (1 - factor) * c2.G;
	c2.B = factor * c1.B + (1 - factor) * c2.B;

	canvas->BeginDraw();

	glBegin(GL_QUAD_STRIP);
	int i;
	CFlatParcoord *test = new CFlatParcoord(flat_info);
	for (i = 0; i < g_globals.dataset->dims; i++) {
		double x = CFlatParcoord::XMAP(i);
		double y1 = CFlatParcoord::YMAP(i, brush->RampMax(i));
		double y2 = CFlatParcoord::YMAP(i, brush->Max(i));

		glColor3f((float) c2.R, (float) c2.G, (float) c2.B);
		glVertex2f((float) x, (float) y1);
		glColor3f((float) c1.R, (float) c1.G, (float) c1.B);
		glVertex2f((float) x, (float) y2);
	}
	glEnd();

	glBegin(GL_QUAD_STRIP);
	for (i = 0; i < g_globals.dataset->dims; i++) {
		double x = CFlatParcoord::XMAP(i);
		double y1 = CFlatParcoord::YMAP(i, brush->RampMin(i));
		double y2 = CFlatParcoord::YMAP(i, brush->Min(i));

		glColor3f((float) c2.R, (float) c2.G, (float) c2.B);
		glVertex2f((float) x, (float) y1);
		glColor3f((float) c1.R, (float) c1.G, (float) c1.B);
		glVertex2f((float) x, (float) y2);
	}
	glEnd();

	canvas->EndDraw();
}


// Draw the brush boundary in the parallel coordinate view.
void CFlatParcoord::DrawParcoordBrushBoundRampOutline(grCanvas *canvas,
													  Brush *brush)
{
	// set the appropriate color
	//
	ARR<Vec2> poly(g_globals.dataset->dims);

	int i;
	for (i = 0; i < g_globals.dataset->dims; i++) {
		poly[i].X = CFlatParcoord::XMAP(i);
		poly[i].Y = CFlatParcoord::YMAP(i, brush->RampMax(i));
	}

	canvas->SetXorMode(True);
	canvas->DrawLines(poly);

	for (i = 0; i < g_globals.dataset->dims; i++) {
		poly[i].X = CFlatParcoord::XMAP(i);
		poly[i].Y = CFlatParcoord::YMAP(i, brush->RampMin(i));
	}

	canvas->DrawLines(poly);
	canvas->SetXorMode(False);
}


//set parameters used for drawing, to be rewritten in subclasses;
void CFlatParcoord::SetParameters()
{
	CFlatDisplay::SetParameters();
	SetCalcVals();

	poly.SetLen(dims);
	if (show_dimcluster_band) {
		mapedX = new double [dims];
		mapedYMean = new double [dims];
		mapedYMax = new double [dims];
		mapedYMin = new double [dims];
		for (int d = 0; d < dims; d++) {
			mapedX[d] = XMAP(d);
		}
	}
}

void CFlatParcoord::FreeSpace()
{
	if (show_dimcluster_band) {
		delete mapedX;
		delete mapedYMean;
		delete mapedYMax;
		delete mapedYMin;
	}
}


//draw axes
void CFlatParcoord::DrawAxes()
{
	if (show_dimcluster_band) {
		DrawParcoordAxesWithBands(canvas);
	} else if (show_two_axes) {
		DrawParcoordAxesWithTwoAxes(canvas);
	} else if (show_wide_axes) {
		DrawParcoordWideAxes(canvas);
	} else {
		DrawParcoordAxes(canvas);
	}
	DrawNominalOnAxes(canvas);
}


//set color and style for unbrushed data
void CFlatParcoord::SetUnbrushedColorAndStyle()
{
	CFlatDisplay:: SetUnbrushedColorAndStyle();
	glLineWidth((float) DSIZE);
	glShadeModel(GL_FLAT);  // Because each line is the same color.
}


//set color and style for average of brushes
void CFlatParcoord::SetAverageColorAndStyle()
{
	CFlatDisplay::SetAverageColorAndStyle();
	glLineWidth(3.0);
}


//draw single data
void CFlatParcoord::DrawSingleData(ARR<double> REF data, int datano)
{
	int j = 0;

	// PAUL: draw_org
	// Move BeginDraw to the outside of this function
//	canvas->BeginDraw();
	glBegin(GL_LINE_STRIP);

	for (j = 0; j < g_globals.dataset->dims; j++) {
		glVertex2f((float) XMAP(j), (float) YMAP(j, data[j]));
	}
	glEnd();
//	canvas->EndDraw();

}


void CFlatParcoord::DrawSingleDataWithTwoAxes(ARR<double> mean,
											  ARR<double> max,
											  ARR<double> min,
											  int datano,
											  bool unbrushed)
{
	glBegin(GL_LINE_STRIP);

	for (int j = 0; j < dims; j++) {
		if (g_globals.dim_reducer->IsLeafDim(j)) {
			glVertex2f((float) XMAP(j), (float) YMAP(j, mean[j]));
		} else {
			glVertex2f((float) XMINMAP(j), (float) YMAP(j, min[j]));
			glVertex2f((float) XMAP(j), (float) YMAP(j, mean[j]));
			glVertex2f((float) XMAXMAP(j), (float) YMAP(j, max[j]));
		}
	}
	glEnd();
}

void CFlatParcoord::DrawSingleDataWithDimClusterExtends(ARR<double> mean,
														ARR<double> max,
														ARR<double> min,
														int datano)
{
	DrawSingleData(mean, datano);

	glShadeModel(GL_SMOOTH);

	int d;
	for (d = 0; d < dims; d++) {
		mapedYMean[d] = YMAP(d, mean[d]);
		mapedYMax[d] = YMAP(d, max[d]);
		mapedYMin[d] = YMAP(d, min[d]);
		if (fabs(1 - g_globals.dim_reduction_info->extent_scaling_fac) > 1e-4) {
			mapedYMax[d] = mapedYMean[d] +
						   g_globals.dim_reduction_info->extent_scaling_fac * (mapedYMax[d] -
																	 mapedYMean[d]);
			mapedYMin[d] = mapedYMean[d] +
						   g_globals.dim_reduction_info->extent_scaling_fac * (mapedYMin[d] -
																	 mapedYMean[d]);
		}
	}

	ARR<Vec2> band(4);
	// Draw the lower band

	Vec2 lo, hi;
	lo.X = mapedX[0];
	lo.Y = mapedYMin[0];

	hi.X = lo.X;
	hi.Y = mapedYMean[0];


	band[0] = lo;
	band[3] = hi;
	for (d = 1; d < dims; d++) {
		lo.X = mapedX[d];
		lo.Y = mapedYMin[d];

		hi.X = lo.X;
		hi.Y = mapedYMean[d];

		band[1] = lo;
		band[2] = hi;
		canvas->DrawTransparentBand(band,
									0.0,
									g_globals.dim_reducer->density,
									canvas->fg_color,
									canvas->fg_color);

		band[0] = band[1];
		band[3] = band[2];
	}

	// Draw the upper band
	lo.X = mapedX[0];
	lo.Y = mapedYMean[0];

	hi.X = lo.X;
	hi.Y = mapedYMax[0];
	band[0] = lo;
	band[3] = hi;

	for (d = 1; d < dims; d++) {
		lo.X = mapedX[d];
		lo.Y = mapedYMean[d];

		hi.X = lo.X;
		hi.Y = mapedYMax[d];

		band[1] = lo;
		band[2] = hi;
		//  	  canvas->SetForeground(interface_color[HIGHLIGHT+1]);

		canvas->DrawTransparentBand(band,
									g_globals.dim_reducer->density,
									0.0,
									canvas->fg_color,
									canvas->fg_color);
		//  	  canvas->SetForeground(interface_color[DATA]);

		band[0] = band[1];
		band[3] = band[2];
	}

	glShadeModel(GL_FLAT);
}

// Draw a single brush.
void CFlatParcoord::DrawSingleBrush(Brush *brush)
{
	canvas->SetForeground(interface_color[BRUSH1 + brush->num]);
	//    canvas->SetForeground( interface_color[BRUSH1+brush->num], 0.1);
	canvas->BeginDraw();
	glBegin(GL_QUAD_STRIP);

	for (int j = 0; j < g_globals.dataset->dims; j++) {
		Vec2 lo, hi;

		// x positions are fixed y positions normalized over axis height

		// top points
		hi.X = CFlatParcoord::XMAP(j);
		hi.Y = CFlatParcoord::YMAP(j, brush->Max(j));

		// bottom points
		lo.X = CFlatParcoord::XMAP(j);
		lo.Y = CFlatParcoord::YMAP(j, brush->Min(j));

		glVertex2f((float) lo.X, (float) lo.Y);
		glVertex2f((float) hi.X, (float) hi.Y);
	}

	glEnd();
	canvas->EndDraw();


	/* 
	//  draw boundary for the brush  
	//  this bulk of code can be used if we have to 
	//  draw brush after data (brush has to be transparent, 
	//  so boundary of the brush is needed.  
		canvas->SetForeground( interface_color[TEXT]);
		ARR<double> data;
		data.SetLen(g_globals.dataset->dims);
		for ( int i = 0; i < g_globals.dataset->dims; i++ )
		{
			data[i] = brush->Max(i);
		}
		DrawSingleData(data, 0);
		for ( i = 0; i < g_globals.dataset->dims; i++ )
		{
			data[i] = brush->Min(i);
		}
		DrawSingleData(data, 0);
		canvas->SetForeground( interface_color[BRUSH1+brush->num]);
	*/

	// draw boundary coverage if necessary
	//
	if (brush->bound_type != Brush::STEP) {
		DrawParcoordBrushBoundRamp(canvas,
								   brush,
								   interface_color[BRUSH1 + brush->num],
								   interface_color[BG]);
	}
}


void CFlatParcoord::InputParcoordResizeBrush(grCanvas *canvas,
											 int sx,
											 int sy)
{
	if (axis < 0 || axis >= g_globals.dataset->dims) {
		return;
	}
	canvas->bypassAll(1);
	Vec2 p = canvas->MapScreenToData(Vec2(sx, sy));

	// Clamp y coordinate to within dimension limits
	double y = CFlatParcoord::YINVMAP(axis, p.Y);
	y = MAX2(y, g_globals.dataset->dim_min[axis]);
	y = MIN2(y, g_globals.dataset->dim_max[axis]);

	canvas->SetXorMode(true);

	double mx = CFlatParcoord::XMAP(axis);

	switch (input_mode) {
	case RESIZE_BRUSH_TOP:
		 {
			double ly = g_globals.curr_brush->Min(axis);
			double hy = MAX2(y, ly);
			p.Y = CFlatParcoord::YMAP(axis, hy);

			g_globals.curr_brush->SetMinMax(axis, ly, hy);

			// Resize top
			//
			if (axis > 0) {
				double lx = CFlatParcoord::XMAP(axis - 1);
				double lhy = CFlatParcoord::YMAP(axis - 1,
												 g_globals.curr_brush->Max(axis - 1));
				canvas->DrawLine((float)lx, (float)lhy, (float)mx, (float)p.Y);
			}

			if (axis < g_globals.dataset->dims - 1) {
				double rx = CFlatParcoord::XMAP(axis + 1);
				double rhy = CFlatParcoord::YMAP(axis + 1,
												 g_globals.curr_brush->Max(axis + 1));
				canvas->DrawLine((float)rx, (float)rhy, (float)mx, (float)p.Y);
			}
		} break;

	case RESIZE_BRUSH_BOTTOM:
		 {
			double hy = g_globals.curr_brush->Max(axis);
			double ly = MIN2(y, hy);
			p.Y = CFlatParcoord::YMAP(axis, ly);

			g_globals.curr_brush->SetMinMax(axis, ly, hy);

			// Resize bottom
			//
			if (axis > 0) {
				double lx = CFlatParcoord::XMAP(axis - 1);
				double lly = CFlatParcoord::YMAP(axis - 1,
												 g_globals.curr_brush->Min(axis - 1));
				canvas->DrawLine((float)lx, (float)lly, (float)mx, (float)p.Y);
			}

			if (axis < g_globals.dataset->dims - 1) {
				double rx = CFlatParcoord::XMAP(axis + 1);
				double rly = CFlatParcoord::YMAP(axis + 1,
												 g_globals.curr_brush->Min(axis + 1));
				canvas->DrawLine((float)rx, (float)rly, (float)mx, (float)p.Y);
			}
		} break;

	default:
		break;
	}

	canvas->SetXorMode(false);
	xor_needs_undo = true;
	canvas->bypassAll(0);
}

//this func is used by InputParcoordMouseMotion
void CFlatParcoord::InputParcoordMoveBrush(grCanvas *canvas, int sx, int sy)
{
	if (axis < 0 || axis >= g_globals.dataset->dims) {
		return;
	}

	Vec2 p = canvas->MapScreenToData(Vec2(sx, sy));

	canvas->SetXorMode(true);

	double mx = CFlatParcoord::XMAP(axis);
	double y = CFlatParcoord::YINVMAP(axis, p.Y);
	y = MIN2(y, g_globals.dataset->dim_max[axis] - g_globals.curr_brush->Size(axis) * .5);
	y = MAX2(y, g_globals.dataset->dim_min[axis] + g_globals.curr_brush->Size(axis) * .5);

	double mly = CFlatParcoord::YMAP(axis, y - g_globals.curr_brush->Size(axis) * .5);
	double mhy = CFlatParcoord::YMAP(axis, y + g_globals.curr_brush->Size(axis) * .5);

	if (axis > 0) {
		double lx = CFlatParcoord::XMAP(axis - 1);
		double lly = CFlatParcoord::YMAP(axis - 1,
										 g_globals.curr_brush->Min(axis - 1));
		double lhy = CFlatParcoord::YMAP(axis - 1,
										 g_globals.curr_brush->Max(axis - 1));
		canvas->DrawLine((float)lx, (float)lly, (float)mx, (float)mly);
		canvas->DrawLine((float)lx, (float)lhy, (float)mx, (float)mhy);
	}

	if (axis < g_globals.dataset->dims - 1) {
		double rx = CFlatParcoord::XMAP(axis + 1);
		double rly = CFlatParcoord::YMAP(axis + 1,
										 g_globals.curr_brush->Min(axis + 1));
		double rhy = CFlatParcoord::YMAP(axis + 1,
										 g_globals.curr_brush->Max(axis + 1));
		canvas->DrawLine((float)rx, (float)rly, (float)mx, (float)mly);
		canvas->DrawLine((float)rx, (float)rhy, (float)mx, (float)mhy);
	}

	canvas->SetXorMode(false);

	g_globals.curr_brush->SetPos(axis, y);
	xor_needs_undo = true;
}

//this func is used by InputParcoordMouseMotion & InputParcoordButtonPress
void CFlatParcoord::InputParcoordMultiSelect(grCanvas *canvas,
											 int sx,
											 int sy)
{
	if (axis < 0 || axis >= g_globals.dataset->dims) {
		return;
	}
	if (!multi_select) {
		return;
	}

	// Find all points within circle distance of mouse
	//
	ARR<double> data(g_globals.dataset->dims);
	for (int i = 0; i < g_globals.dataset->data_size; i++) {
		g_globals.dataset->GetData(data, i);

		double y = YMAP(axis, data[axis]);

		Vec2 p = canvas->MapDataToScreen(Vec2(XMAP(axis), y));

		if (fabs(p.Y - sy) < multisel_radius) {
			multiSelectAddPoint(i);

			glDrawBuffer(GL_FRONT);
			canvas->SetForeground(interface_color[PAINT]);
			canvas->bypassAll(1);
			DrawSingleData(data);
			canvas->bypassAll(0);
			glDrawBuffer(GL_BACK);
		}
	}
}

//this func is used by InputParcoordMouseMotion & InputParcoordButtonPress
void CFlatParcoord::InputParcoordBoundRampResize(grCanvas *canvas,
												 int sx,
												 int sy)
{
	if (axis < 0 || axis >= g_globals.dataset->dims) {
		return;
	}

	Vec2 p = canvas->MapScreenToData(Vec2(sx, sy));

	double y = CFlatParcoord::YINVMAP(axis, p.Y);

	if (y > g_globals.curr_brush->Pos(axis)) {
		g_globals.curr_brush->SetBoundOffset(axis, y - g_globals.curr_brush->Max(axis));
		g_globals.curr_brush->SetBoundOffset(axis, MAX2(g_globals.curr_brush->BoundOffset(axis), 0.0));
	} else {
		g_globals.curr_brush->SetBoundOffset(axis, g_globals.curr_brush->Min(axis) - y);
		g_globals.curr_brush->SetBoundOffset(axis, MAX2(g_globals.curr_brush->BoundOffset(axis), 0.0));
	}

	DrawParcoordBrushBoundRampOutline(canvas, g_globals.curr_brush);

	xor_needs_undo = true;
}


void CFlatParcoord::InputMouseMotion(int x, int y)
{
	if (!g_globals.dataset) {
		return;
	}

	if (g_globals.screen_distortion->distortion_status) {
		Vec2 p = canvas->MapScreenToData(Vec2(x, y));
		InputParacoordFindAxesBetween(canvas, p.X, p.Y, axis);
		return;
	}

	ParcoordCaption(canvas, x, y);
	static int prev_sx, prev_sy;

	switch (input_mode) {
	case CANVAS_PAN:
		appCanvasMousePan(canvas);
		return;

	case RESIZE_RAMP:
		if (xor_needs_undo)
			InputParcoordBoundRampResize(canvas, prev_sx, prev_sy);
		InputParcoordBoundRampResize(canvas, x, y);
		break;

	case RESIZE_BRUSH_TOP:
	case RESIZE_BRUSH_BOTTOM:
		if (xor_needs_undo)
			InputParcoordResizeBrush(canvas, prev_sx, prev_sy);
		InputParcoordResizeBrush(canvas, x, y);
		break;

	case TRANSLATE_BRUSH:
		if (xor_needs_undo)
			InputParcoordMoveBrush(canvas, prev_sx, prev_sy);
		InputParcoordMoveBrush(canvas, x, y);
		break;

	case MULTI_SELECT:
		InputParcoordMultiSelect(canvas, x, y);
		return;

	default:
		break;
	}

	prev_sx = x;
	prev_sy = y;
}


void CFlatParcoord::InputButtonPress(int button, int x, int y)
{
	if (!g_globals.dataset) {
		return;
	}

	input_mode = NONE;

	if ((!canvas->altkey) &&
		(!canvas->controlkey) &&
		canvas->button[2] == true) {
		input_mode = CANVAS_PAN;
		canvas->SetHandCursor();
		return;
	}  

	Vec2 p = canvas->MapScreenToData(Vec2(x, y));
	InputParacoordFindAxesNear(canvas, p.X, p.Y, axis);

	if (axis < 0 || axis >= g_globals.dataset->dims) {
		axis = -1;
		return;
	}

	if (canvas->shiftkey && button == 0) {
		input_mode = MULTI_SELECT;
		InputParcoordMultiSelect(canvas, x, y);
		return;
	} else if (canvas->controlkey &&
			   g_globals.curr_brush->bound_type == Brush::RAMP &&
			   button == 0) {
		input_mode = RESIZE_RAMP;
		InputParcoordBoundRampResize(canvas, x, y);
		return;
	}

	assert(xor_needs_undo == false);

	if (canvas->button[0] == true) {
		double lo = CFlatParcoord::YMAP(axis, g_globals.curr_brush->Min(axis));
		double hi = CFlatParcoord::YMAP(axis, g_globals.curr_brush->Max(axis));
		if ((hi - p.Y) > (p.Y - lo)) {
			input_mode = RESIZE_BRUSH_TOP;
		} else {
			input_mode = RESIZE_BRUSH_BOTTOM;
		}
	} else if (canvas->button[1] == true) {
		input_mode = TRANSLATE_BRUSH;
	}

#ifdef USE_META
	if (canvas->controlkey && button == 2) {
		int datax = CLAMP(axis, 0, g_globals.dataset->dims - 1);
		double datay = YINVMAP(datax, p.Y);
		g_globals.dataset->ShowContents(Togl_Interp(canvas->GetWidget()),
							  datax,
							  datay);
	}
#endif // USE_META

	InputMouseMotion(x, y);
}


void CFlatParcoord::InputButtonRelease(int button, int x, int y)
{
	if (!g_globals.dataset) {
		return;
	}

	//try
	if (canvas->altkey && button == 2 && flat_info->flat_sb_brush) {
		HierChangeFlatBrushAccordingToClusters();
		canvas->Refresh();
	}

	if (g_globals.screen_distortion->distortion_status) {
		double delta;

		input_mode = INCREASE_AXES_INTERVAL;
		switch (input_mode) {
		case INCREASE_AXES_INTERVAL:
			delta = 1. / g_globals.dataset->dims * 0.2;
			InputParacoordScaleAxes(canvas, axis, delta);
			canvas->Refresh();
			break;
		case DECREASE_AXES_INTERVAL:
			delta = (-1.) / g_globals.dataset->dims * 0.2;
			InputParacoordScaleAxes(canvas, axis, delta);
			break;
		default:
			break;
		}
		return;
	}

	if (canvas->controlkey &&
		button == 2 &&
		g_globals.dim_reducer != NULL &&
		g_globals.dim_reducer->status == APPLY) {
		Vec2 p = canvas->MapScreenToData(Vec2(x, y));
		InputParacoordFindAxesNear(canvas, p.X, p.Y, axis);

		if (axis < 0 || axis >= g_globals.dataset->dims) {
			axis = -1;
			return;
		}

		if (!canvas->altkey)
			g_globals.dim_reducer->ExtendNodeInRepresentativeDimList(axis);
		else
			g_globals.dim_reducer->ShinkNodeInRepresentativeDimList(axis);

		appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
	}

	switch (input_mode) {
	case CANVAS_PAN:
		canvas->SetDefaultCursor();
		break;

	case MULTI_SELECT:
		break;

	case RESIZE_RAMP:
	case RESIZE_BRUSH_TOP:
	case RESIZE_BRUSH_BOTTOM:
	case TRANSLATE_BRUSH:
		BrushOperMarkDataset();
		canvas->SetDefaultCursor();
		appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
		break;

	default:
		break;
	}

	input_mode = NONE;
	xor_needs_undo = false;
}

bool CFlatParcoord::TestDimensionality()
{
	return ParcoordTestDimensionality();
}

void CFlatParcoord::ClutterBasedOrder()	// added by debbie 9/2003
{
	g_globals.dataset->Clutter_Based_Order("par");  
}
