///////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool - A program for Multivariate Data Display
//
//  Copyright 1994-2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products	
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
///////////////////////////////////////////////////////////// COPYRIGHT //

//////////////////////////////////////////////////////////////////////////
/// \file CHierNavhHelperDirectDb.h
/// \brief This file implements the class CHierNavHelper.
/// 
/// This class evaluates the tree_contour required by the class CHierNav 
/// using the values from the data base. This class is instantiated only 
/// when the tool is connected to the data base
///
/// \date 09\25\2003
/// \author Nishant Mehta
//////////////////////////////////////////////////////////////////////////

/*============================== INCLUDES ==============================*/

#include <xmdv/CHierNavHelperDirectDb.h>		// This file has the super class from which this file inherits
#ifdef XMDV_USE_BACKEND
#include <DataBaseLibrary/ResultSet.h>
#include <DataBaseLibrary/Statement.h>
#include <sstream>
#include <string>


	
	CHierNavHelperDirectDb::CHierNavHelperDirectDb(Statement *statement,
												   std::string sourceTableName,
												   std::string targetTableName)
	{
		this->statement = statement;
		this->srcTableName = sourceTableName;
		this->tableName = targetTableName;
	}

	/// Destructor
	CHierNavHelperDirectDb::~CHierNavHelperDirectDb()
	{}

	
	///
	/// Find the radius of the root node
	/// 
	/// \return the radius of the root node
	///
	double	CHierNavHelperDirectDb :: FindRootRadius()
	{		
		std::string query = "select max(RADIUS) from "+tableName;		
		ResultSet* resultSet = statement->executeQuery(query);
		double radius = 0;
		while(!resultSet->isDone())
		{
			resultSet->next();
			radius = resultSet->getDouble(0);
		}
		delete(resultSet);
		return(radius);
	}

	

	///
	/// Populates the tree_contour array
	///
	/// This function evaluates fetches all the different levels from the data base
	/// That is the height of the tree_contour is the same as 
	/// as that of the length of this array. 
	///
	/// \param countour_radius an array that contains all the distinct values 
	///					       for the radius
	///
	void CHierNavHelperDirectDb ::FindRadiusDistribution(ARR<double> &contour_radius)
	{
	
		std::string query = "select count(*) from "+tableName;
		
		ResultSet*  resultSet = statement->executeQuery(query);
		while(!resultSet->isDone())
		{
			resultSet->next();
			int len = (int) resultSet->getDouble(0);
			contour_radius.SetLen(len);
		}
		delete(resultSet);
		query = "select RADIUS from "+tableName+" order by RADIUS desc ";		
		resultSet = statement->executeQuery(query);

		double radius = 0;
		int i = 0;
		while(!resultSet->isDone())
		{
			resultSet->next();
			contour_radius[i] = resultSet->getDouble(0);
			i++;
		}		
		delete(resultSet);
	}

	///
	/// Populates the tree_contour
	///
	/// This function evaluates the complete tree_contour. 
	/// That is for each level ot fetches all the nodes at the levels and then
	/// sets the values in the tree_contour[i]. Where i corresponds to the depth 
	/// of the tree you are currently working at
	/// 
	/// \param &tree_contour a 2 dimensional array that stores the complete tree_contour
	/// \param &tree_color a 2 dimensional array that stores the color
	/// \param contour_radius this is used to evaluate the contour, gives all  different level
	///
	void CHierNavHelperDirectDb ::	FindTreeContour(ARR<ARR<double> > &tree_contour,
							ARR<ARR<double> > &tree_color,
							ARR<double> contour_radius)
	{
		std::ostringstream queryStream;
		queryStream<<" select * from "
				    <<this->tableName
					<<" order by RADIUS desc ";
		
		ResultSet* resultSet = statement->executeQuery(queryStream.str());
		int i = 0;
		while(!resultSet->isDone())
		{
			
			resultSet->next();
			contour_radius[i] = resultSet->getDouble(0);
			//Assigning the tree contour
			for(int j = 0 ; j< CONTOUR_RES;j++)
				tree_contour[i][j] = resultSet->getDouble(j+1);
			//Assigning the tree color
			for(int j = 0;j< CONTOUR_RES;j++)
				tree_color[i][j] = resultSet->getDouble(j+CONTOUR_RES+1);

			i++;
		}
		delete(resultSet);
	}

	/*
	* Returns the height of the tree
	*/
	int	CHierNavHelperDirectDb::FindTreeHeight()
	{
		std::string query = "select max(height) from " + this->srcTableName;
		ResultSet* resultSet = statement->executeQuery(query);
		int height;
		while(!resultSet->isDone())
		{
			resultSet->next();
			height = (int) resultSet->getDouble(0);
		}
		delete(resultSet);
		return(height);
	}

#endif //XMDV_USE_BACKEND
