#include <xmdv/CHierTreemap.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/cluster.h>
#include <xmdv/colormap.h>
#include <xmdv/ogl.h>
#include <xmdv/okcdata.h>
#include <xmdv/colorramp.h>

int dep_dim	= 0;

static double max_val, min_val;

CHierTreemap::CHierTreemap(CHierDisplayInformation *hier_info)
	: CHierDisplay(hier_info)
{
	treemap_CMAP = new colormap("TreeMap",
								ColorRamp::GetPredefinedGreyScale(CMAPSIZE));
}

CHierTreemap::~CHierTreemap()
{
	delete treemap_CMAP;
}

// Allocate a rectangle for each node in the treemap
//
static void cutTreeMap(Cluster *node, double L, double R, double B,
					   double T, int curr_ht)
{
	node->xmin = L;
	node->ymin = B;

	node->xmax = R;
	node->ymax = T;

	if (node->num_children > 0)
	{
		//    if (node->entries > g_globals.btree->treemap_limit_number) {
		// Cut vertically if odd-level and horizontally if even-level
		//
		double k, color_l, color_r;

		if (curr_ht % 2)
		{
			color_l = L;
			color_r = L;
			for (int i = 0; i < node->num_children; i++)
			{
				k = (R - L) * ((double)
							   g_globals.btree->nodes[node->child[i]].entries /
							   (double)
							   node->entries);
				color_l = color_r;
				color_r = color_l + k;
				cutTreeMap(&g_globals.btree->nodes[node->child[i]],
						   color_l, color_r, B, T, curr_ht + 1);
			}
		} else
		{
			color_l = B;
			color_r = B;
			for (int i = 0; i < node->num_children; i++)
			{
				k = (T - B) * ((double)
							   g_globals.btree->nodes[node->child[i]].entries /
							   (double)
							   node->entries);
				color_l = color_r;
				color_r = color_l + k;
				cutTreeMap(&g_globals.btree->nodes[node->child[i]], L, R,
						   color_l, color_r, curr_ht + 1);
			}
		}
	}
}



static void getMinMaxLeafDimension(Cluster *node)
{
	if (node->num_children == 0)
	{
		if (node->sx[dep_dim] < min_val)
			min_val = node->sx[dep_dim];
		if (node->sx[dep_dim] > max_val)
			max_val = node->sx[dep_dim];
	}

	for (int i = 0; i < node->num_children; i++)
		getMinMaxLeafDimension(&g_globals.btree->nodes[node->child[i]]);
}

void cfTreeMapInit()
{
	if (!g_globals.btree)
		return;

	max_val = -HUGE_VAL;
	min_val = HUGE_VAL;
	getMinMaxLeafDimension(&g_globals.btree->nodes[0]);
}

#define BORDER 10
#define TOP_BORDER BORDER
#define BOT_BORDER 5*BORDER
#define BPARA 0.9

static double TransformYInTreemap(double y)
{
	return y * BPARA;
}


void CHierTreemap::paintLeaves(grCanvas *canvas, Cluster *node)
{
	double range= max_val - min_val;
	double color= 0.0;

	if (node->num_children == 0) 
		//    if (node->entries <= g_globals.btree->treemap_limit_number)
	{
		if (BITTEST(node->flags, IS_BRUSHED))
		{
			if (range > 0.0)
				// Highlight based on one dimension's value
				color = (node->sx[dep_dim] / node->entries - min_val) /
						range;
			assert(color >= 0 && color <= 1);
			canvas->SetForeground((*treemap_CMAP)[(int)
								  (color * (CMAPSIZE - 1))]);
		} else
		{
			// Consider color as one of the dimension
			canvas->SetForeground((*g_globals.hier_info->CMAP)[(int)
								  (node->color * (CMAPSIZE - 1))]);
		}

		canvas->FillRectangle(node->xmin, TransformYInTreemap(node->ymin),
							  node->xmax - node->xmin,
							  TransformYInTreemap(node->ymax - node->ymin));
		return;
	} else
		for (int i = 0; i < node->num_children; i++)
			paintLeaves(canvas, &g_globals.btree->nodes[node->child[i]]);
}

void CHierTreemap::traverseTree(grCanvas *canvas, Cluster *node,
								int curr_ht, double radius, bool brushed)
{
	// Root node, we want to trace the rectangle outline
	//
	if (curr_ht == 0)
		canvas->DrawRectangle(node->xmin, TransformYInTreemap(node->ymin),
							  node->xmax - node->xmin,
							  TransformYInTreemap(node->ymax - node->ymin));

	if (curr_ht > 0 &&
		g_globals.btree->nodes[node->parent].sqrt_radius <= radius)
		return;

	if (BITTEST(node->flags, IS_BRUSHED) == brushed)
	{
		// Draw the dividing lines at each level
		if (curr_ht % 2)
			canvas->DrawLine((float)node->xmax, 
							(float)TransformYInTreemap(node->ymin),
							(float)node->xmax, 
							(float)TransformYInTreemap(node->ymax));
		else
			canvas->DrawLine((float)node->xmin, 
							(float)TransformYInTreemap(node->ymax),
							(float)node->xmax, 
							(float)TransformYInTreemap(node->ymax));
	}

	if (node->num_children > 0) 
		//    if (node->entries > g_globals.btree->treemap_limit_number)
		for (int i = 0; i < node->num_children; i++)
			traverseTree(canvas, &g_globals.btree->nodes[node->child[i]],
						 curr_ht + 1, radius, brushed);
}


void CHierTreemap::Draw()
{
	if (!g_globals.dataset || !g_globals.btree)
		return;

	PrepareCanvas();

	cutTreeMap(&g_globals.btree->nodes[0], 0, 1, 0, 1, 1);

	canvas->bypassAll(true);
	paintLeaves(canvas, &g_globals.btree->nodes[0]);

	double r= g_globals.btree->nodes[0].sqrt_radius;


	//canvas->SetForeground(0x000000);

	traverseTree(canvas, &g_globals.btree->nodes[0], 0,
				 g_globals.hier_info->cluster_radius * r, 0);
	traverseTree(canvas, &g_globals.btree->nodes[0], 0,
				 g_globals.hier_info->brushed_radius * r, 1);

	canvas->bypassAll(false);
	//    glViewport( BORDER, BORDER, W-2*BORDER, 2*BORDER);

	// Write the dependent dimension
	//
	char str[50];
	sprintf(str, "Dependent Dimension is %s",
			g_globals.dataset->names[dep_dim]);
	canvas->SetForeground(0x000000);
	canvas->DrawString(0, BPARA + (1 - BPARA) * 0.4, str);

	strcpy(str, "Low");
	canvas->DrawString(0, BPARA + (1 - BPARA) * 0.7, str);

	strcpy(str, "High");
	canvas->DrawString(0.9, BPARA + (1 - BPARA) * 0.7, str);

	//
	// Draw color bar
	//
	glShadeModel(GL_SMOOTH);
	glLineWidth(float(8.0 * canvas->zoom.Y));
	glBegin(GL_LINE_STRIP);
	int cmap_size	= CMAPSIZE;
	double step		= 1 / double(cmap_size - 1);
	for (int i = 0; i < cmap_size; i++)
	{
		unsigned long color = (*treemap_CMAP)[i];
		glColor3ub((unsigned char) (color >> 16) & 0xff,
				   (unsigned char) (color >> 8) & 0xff,
				   (unsigned char) (color) & 0xff);
		double k= i *step;
		glVertex2f((float) k, 1);
	}
	glEnd();
	glLineWidth(1.0);
}


/*
void CHierTreemap::Draw()
{
	if (!g_globals.dataset || !g_globals.btree) return;

	canvas->MakeCurrent();
	canvas->SetTopLeftOrigin();
	grCanvas::grResizeMainCanvas(canvas);


	//
	// Setup window parameters.
	//
	int W, H;
	canvas->GetSize(&W,&H);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glViewport( BORDER, BOT_BORDER, W-2*BORDER, H-BOT_BORDER-TOP_BORDER);
	glOrtho(0, 1, 0, 1, -1,1);

	canvas->SetBackground(0x000000);
	canvas->Clear();

	cutTreeMap(&g_globals.btree->nodes[0], 0, 1, 0, 1, 1);

	paintLeaves(canvas, &g_globals.btree->nodes[0]);

	double r = g_globals.btree->nodes[0].sqrt_radius;

	canvas->SetForeground(0x000000);

	traverseTree(canvas, &g_globals.btree->nodes[0], 0, g_globals.hier_info->cluster_radius*r, 0);
	traverseTree(canvas, &g_globals.btree->nodes[0], 0, g_globals.hier_info->brushed_radius*r, 1);

	glViewport( BORDER, BORDER, W-2*BORDER, 2*BORDER);

	// Write the dependent dimension
	//
	char str[50];
	sprintf(str,"Dependent Dimension is %s", g_globals.dataset->names[dep_dim]);
	canvas->SetForeground(0xffffff);
	canvas->DrawString(0,1,str);

	strcpy(str,"Low");
	canvas->DrawString(0,0.4,str);

	strcpy(str,"High");
	canvas->DrawString(0.95,0.4,str);

	treemap_CMAP.DrawColorBar();

}
*/
