// Display data using parallel coordinates.  Each dimensional axis is layed
// out in a vertical line, evenly spaced between dimensions.  The top and
// bottom of each line coincide with the maximum and minimum for that
// dimension.  Each data point creates a polyline across these axes.  Labels
// are printed along the top of each dimension, and brush coverage display
// consists of a single shaded region across the drawing area.

#include <xmdv/CParcoord.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/ogl.h>
#include <xmdv/CDimReducer.h>
#include <xmdv/CScreenDistortion.h>

#ifdef USE_META
#include <meta/Meta.h>
#endif // USE_META

CParcoord::CParcoord()
{
}


CParcoord::~CParcoord()
{
}


bool CParcoord::ParcoordTestDimensionality()
{
	if ( g_globals.dataset->dims < 2 )
		return false;
	else 
		return true;
}

// Draw parallel coordinate axes.
void CParcoord::DrawParcoordAxes(grCanvas *canvas)
{
    glLineWidth(1.);
    glShadeModel(GL_FLAT);
	
    // Draw the vertical axes
    int i;
    for (i=0; i<g_globals.dataset->dims; i++) {
        canvas->SetForeground(g_globals.dataset->axis_color[i]);
		canvas->BeginDraw();
        double x = XMAP(i);
		
        canvas->DrawLine((float)x, 0.0f, (float)x, 1.0f);
		canvas->EndDraw();
    }
	canvas->Flush();
    //Draw axes names and mins and maxes
    DrawParcoordAxesNameMinMax(canvas);
}

// Draw parallel coordinate axes.
void CParcoord::DrawParcoordAxesWithBands(grCanvas *canvas)
{
    glLineWidth(1.);
    glShadeModel(GL_SMOOTH);

    // Draw the vertical axes
    int i;
    for (i=0; i<g_globals.dataset->dims; i++) {
        canvas->SetForeground(g_globals.dataset->axis_color[i]);
        double x = XMAP(i);
        canvas->DrawLine((float)x, 0.0f, (float)x, 1.0f);
    }
	canvas->Flush();

    //draw name, min and max
    DrawParcoordDimensionClusterAxesNameMinMax(canvas);
}

// Draw parallel coordinate axes.
void CParcoord::DrawParcoordWideAxes(grCanvas *canvas)
{
    glShadeModel(GL_SMOOTH);
	glLineWidth(1.);
    // Draw the vertical axes
    //
    int i;
    for (i=0; i<g_globals.dataset->dims; i++) {
        canvas->SetForeground(g_globals.dataset->axis_color[i]);
        /* set width of the axis */
        if ( !g_globals.dim_reducer->IsLeafDim(i) )
        {
			canvas->bypass(1);
            double dim_disagreement = g_globals.dim_reducer->GetDimDisagreement(i);
            glLineWidth((float)(1.+ dim_disagreement * 10.));
        }
        double x = XMAP(i);
        canvas->DrawLine((float)x, 0.0f, (float)x, 1.0f);
    }
	canvas->Flush();

    //draw name, min and max
    DrawParcoordDimensionClusterAxesNameMinMax(canvas);
}

// Draw standard vertical uniformly spaced parallel coordinate axes, with two axes to show disagree of dim cluster.
void CParcoord::DrawParcoordAxesWithTwoAxes(grCanvas *canvas)
{
    glLineWidth(1.);
    glShadeModel(GL_SMOOTH);

    // Draw the mean vertical axes
    //
    int i;
    for (i=0; i<g_globals.dataset->dims; i++) 
    {
        canvas->SetForeground(g_globals.dataset->axis_color[i]);
        double x = XMAP(i);
        canvas->DrawLine((float)x, 0.0f, (float)x, 1.0f);
    }

    // Draw the min and max vertical axes
    //
    for (i=0; i<g_globals.dataset->dims; i++) 
    {
        if ( !g_globals.dim_reducer->IsLeafDim(i) )
        {
            canvas->SetForeground(interface_color[GRID2]);
            double x = XMINMAP(i);
            canvas->DrawLine((float)x, 0.0f, (float)x, 1.0f);

            canvas->SetForeground(interface_color[GRID3]);
            x = XMAXMAP(i);
            canvas->DrawLine((float)x, 0.0f, (float)x, 1.0f);
        }
    }
	canvas->Flush();
    //draw name, min and max
    DrawParcoordDimensionClusterAxesNameMinMax(canvas);
}


void CParcoord::InputParacoordFindAxesBetween(grCanvas *canvas, double X, double Y, int &dim)
{
    if ( Y < 0 || Y > 1.0 )
    {
        dim = -1;
    }
    else
    {
        int i = -1;
        while ( i < g_globals.dataset->dims-1 && X > g_globals.screen_distortion->paracoord_axes[i+1] )
            i++;

        dim = i;
    }
    return;
}


void CParcoord::InputParacoordFindAxesNear(grCanvas *canvas, double X, double Y, int &dim)
{
    if ( Y < 0. || Y > 1.0 )
    {
        dim = -1;
        return;
    }
    else
    {
        if ( X < 0. || X > 1.0 )
        {
            dim = -1;
            return;
        }

        for ( int i = 0; i < g_globals.dataset->dims-1; i++ )
        {
            if ( X < ( g_globals.screen_distortion->paracoord_axes[i] + g_globals.screen_distortion->paracoord_interval[i]/2 ) )
            {
                dim = i;
                return;
            }
        }

        dim = g_globals.dataset->dims-1;
        return;
    }
}


//this func is used by InputParcoordMouseMotion
void CParcoord::ParcoordCaption(grCanvas *canvas, int x, int y)
{
    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );

//    int datax = int(p.X * g_globals.dataset->dims);
//    datax = CLAMP(datax,0,g_globals.dataset->dims-1);
    int datax;
    
    InputParacoordFindAxesNear(canvas, p.X, p.Y, datax);

    datax = CLAMP(datax,0,g_globals.dataset->dims-1);
    
    double datay = YINVMAP(datax,p.Y);

    char caption[100];
	sprintf(caption,"%s=",g_globals.dataset->names[datax]);
	strcat(caption, g_globals.dataset->MapValueToString(datax, datay));
	
    appShowCanvasMessage(Togl_Interp(canvas->GetWidget()),caption, canvas->name);
}


void CParcoord::InputParacoordScaleAxes(grCanvas *canvas, int dim, double delta)
{
    if ( dim < 0 || dim > g_globals.dataset->dims )
        return;

    g_globals.screen_distortion->ParacoordScaleAxes(dim, delta);
    return;
}

// PAUL: lang_opt
//	This is now inline
double CParcoord::XMAP(int I)
{ 
    return g_globals.screen_distortion->paracoord_axes[I];
}


double CParcoord::XMINMAP(int I)
{ 
    return g_globals.screen_distortion->paracoord_axes[I]-0.04;
}

double CParcoord::XMAXMAP(int I)
{ 
    return g_globals.screen_distortion->paracoord_axes[I]+0.04;
}


double CParcoord::YMAP(int X, double Y)
{
//    return 1.0-(Y - g_globals.dataset->dim_min[X]) / (g_globals.dataset->dim_max[X] - g_globals.dataset->dim_min[X]);
	// PAUL: preorder_math
    return 1.0-(Y - g_globals.dataset->dim_min[X]) * inv_dim_range[X];	
}


//it is for the uniformed dim: 0-1
double CParcoord::YMAP(double Y)
{
    return 1.0-Y;
}


double CParcoord::YINVMAP(int X, double Y)
{
    return ((1.0-(Y)) * (g_globals.dataset->dim_max[X] - g_globals.dataset->dim_min[X])) + g_globals.dataset->dim_min[X];
}

// Draw parallel coordinate axes name and min max.
void CParcoord::DrawParcoordAxesNameMinMax(grCanvas *canvas)
{
    glShadeModel(GL_SMOOTH);
    canvas->SetForeground(interface_color[TEXT]);

    Vec2 font_size = canvas->GetDataFontSize("W");

	double max_str_size;

    for (int i=0; i<g_globals.dataset->dims; i++)
    {
        if ( i != g_globals.dataset->dims-1 )
		    max_str_size = g_globals.screen_distortion->paracoord_interval[i] - font_size.X*5; 
        else 
		    max_str_size = g_globals.screen_distortion->paracoord_interval[i]; 
        double x = g_globals.screen_distortion->paracoord_axes[i];
        canvas->DrawConfinedString(x,font_size.Y,g_globals.dataset->names[i], max_str_size);

        // Draw the MAX2 values of the dimension
		canvas->DrawConfinedString(x,-.5*font_size.Y,g_globals.dataset->GetMax(i),max_str_size);
		
        // Draw the MIN2 values of the dimension
        canvas->DrawConfinedString(x,1.0-font_size.Y,g_globals.dataset->GetMin(i),max_str_size);
    }
}

// Draw parallel coordinate axes name and min max in dissimilarity displays.
void CParcoord::DrawParcoordDimensionClusterAxesNameMinMax(grCanvas *canvas)
{
    glShadeModel(GL_SMOOTH);
    canvas->SetForeground(interface_color[TEXT]);

    Vec2 font_size = canvas->GetDataFontSize("W");

	double max_str_size;

    for (int i=0; i<g_globals.dataset->dims; i++)
    {
		max_str_size = g_globals.screen_distortion->paracoord_interval[i] - font_size.X*4; 
        double x = g_globals.screen_distortion->paracoord_axes[i];
        char name[50], name_a[30];
        strcpy(name, g_globals.dataset->names[i]);
        sprintf(name_a, "(%1d)", g_globals.dim_reducer->GetEntries(i));
        strcat(name, name_a);
        canvas->DrawConfinedString(x,font_size.Y, name, max_str_size);

        // Draw the MAX2 values of the dimension
        canvas->DrawConfinedString(x,-.5*font_size.Y,g_globals.dataset->GetMax(i),max_str_size);

        // Draw the MIN2 values of the dimension
        canvas->DrawConfinedString(x,1.0-font_size.Y,g_globals.dataset->GetMin(i),max_str_size);
    }
}

// PAUL: precalc_math
void CParcoord::SetCalcVals()
{
	// Divisions are expensive, so do the minimum number of divisions here before drawing
	inv_dim_range.SetLen(g_globals.dataset->dims);

	for (int i = 0; i < g_globals.dataset->dims; ++i)
		inv_dim_range[i] = 1 / (g_globals.dataset->dim_max[i] - g_globals.dataset->dim_min[i]);
}

//Draw the nominal value beside the axes
void CParcoord::DrawNominalOnAxes(grCanvas *canvas)
{
#ifdef USE_META

		Vec2 font_size = canvas->GetDataFontSize("W");
			double max_str_size;
		//if ( j != g_globals.dataset->dims-1 )
		//	max_str_size = g_globals.screen_distortion->paracoord_interval[j] - font_size.X*5; 
		//else 
			max_str_size = g_globals.screen_distortion->paracoord_interval[0]; 
			for (int j=0; j<g_globals.dataset->dims; ++j) {
				IMeta *meta = g_globals.dataset->GetMeta();
				if (meta) {
					Type dimType;
					
					int jReal = g_globals.dataset->GetRealPos(j);
					IDimension *dimension = meta->getDimension(jReal);
					if (dimension) {
							dimType = dimension->getType();
					}
					// this should be a valid type
					assert(!dimType.isBad());
					
					if (dimType.isDerivedFrom(Nominal::getTypeId())) 
					{
						IDimension::MappingIterator *mapIter = dimension->getMappingIterator();
						while (mapIter->hasNext()) {
							float fValue; IValue *mappedValue;
							mapIter->next(&fValue, &mappedValue);
							
							float x = (float) XMAP(j);
							float y = (float) YMAP(j,fValue);
							canvas->SetForeground(interface_color[TEXT]);
							char s[255];
							strcpy(s, mappedValue->getStringContents());
							canvas->DrawConfinedString(x,y,mappedValue->getStringContents(), max_str_size);
						}
						
					}
				}
			}
			
#endif // USE_META
}


