///////////////////////////////////////////////////////////////////////////////
/// \file Subwindow.cpp
/// \brief Implementation of the Subwindow class
/// \date 5/6/2004
/// \author Anilkumar Patro
///////////////////////////////////////////////////////////////////////////////

// SYSTEM INCLUDES
//

// PROJECT INCLUDES
//
#include <xmdv/Subwindow.h>
#include <xmdv/Image.h>
#include <xmdv/colormap.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/CHierDisplayInformation.h>

namespace
{
	double samplingFactor;
};

// FORWARD REFERENCES
//
void sspiral(double startSample, double endSample, Image *rgb, int dim,
			 const QueryDist &query_dist, CHierDisplayInformation *hier_info = NULL);

//-----------------------------------------------------------------------------
Subwindow::Subwindow(int dim)
	: m_box()
{
	m_dimension = dim;
	m_size = -1;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
Subwindow::Subwindow(const Subwindow& from)
	: m_box(from.m_box)
{
	m_dimension = from.m_dimension;
	m_size = from.m_size;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
 Subwindow::~Subwindow()
{
	// TODO: Add destructor
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
Subwindow& Subwindow::operator = (const Subwindow& from)
{
	if (this == &from)
		return *this;

	m_box = from.m_box;
	m_dimension = from.m_dimension;
	m_size = from.m_size;

	return *this;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
void Subwindow::Paint(Image *img,const QueryDist& flat_query, 
					  double startSample /*=0*/,double endSample /*=1.0*/,
					  bool hilite /*=false*/)
{
	assert(m_box.IsValid());

	// hilite dimension if it's the current one
	if (hilite) {
		glLineWidth(5.0f);
		RGBt color(interface_color[TEXT]);
		glColor3f((float) color.R, (float) color.G, (float) color.B);
		glBegin(GL_LINE_LOOP); {
			glVertex2d(m_box.Lowx(),  m_box.Lowy());
			glVertex2d(m_box.Highx(), m_box.Lowy());
			glVertex2d(m_box.Highx(), m_box.Highy());
			glVertex2d(m_box.Lowx(),  m_box.Highy());
		} glEnd();
		glLineWidth(1.0f);
	}

	// draw the spiral arrangement of pixels on the subwindow image
	m_size = img->Width()-1;
	sspiral(startSample, endSample, img, m_dimension, flat_query);
	img->CreateTexture();
	img->Display((float) m_box.Lowx(), (float) m_box.Highy(), (float) m_box.Highx(), (float) m_box.Lowy());
	img->DeleteTexture();
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
void Subwindow::PaintClusters(Image *img, int size, const QueryDist& hier_query, 
							  CHierDisplayInformation *hier_info,
							  bool hilite /* = false */)
{
	assert(m_box.IsValid());

	// hilite dimension if it's the current one
	if (hilite) {
		glLineWidth(5.0f);
		glColor3f(0,0,0);
		glBegin(GL_LINE_LOOP); {
			glVertex2d(m_box.Lowx(),  m_box.Lowy());
			glVertex2d(m_box.Highx(), m_box.Lowy());
			glVertex2d(m_box.Highx(), m_box.Highy());
			glVertex2d(m_box.Lowx(),  m_box.Highy());
		} glEnd();
		glLineWidth(1.0f);
	}
	samplingFactor = 1.0f;
	m_size = size;

	// draw the spiral arrangement of pixels on the subwindow image
//	sspiralClusters(m_size/2, m_size/2, m_size, img, cmap, m_dimension, 
//					infos, hier_info);
	sspiral(0, 1.0, img, m_dimension, 
			hier_query, hier_info);
	img->CreateTexture();
	img->Display((float) m_box.Lowx(), (float) m_box.Highy(), (float) m_box.Highx(), (float) m_box.Lowy());
	img->DeleteTexture();
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
bool Subwindow::Contains(float xx, float yy, int &n)
{
	if (m_box.IsInside(xx, yy) && m_size >= 0)
	{
		double xRatio = (xx - m_box.Lowx()) / m_box.Width();
		double yRatio = (m_box.Highy() - yy) / m_box.Height();

		int x = (int) (m_size * xRatio), y = (int) (m_size * yRatio);
		x = x - m_size/2; y = y - m_size/2;

		// from: http://mcraefamily.com/MathHelp/CountingRationalsSquareSpiral1.htm
		//
		int shell = 0, leg = 0, elem = 0;

		//
		// Given (x,y),
		//
		// shell -- is the maximum of the distance in either dimension from the 
		// origin, in other words: max(abs(x),abs(y))
		//
		//shell = max(abs(x), abs(y));
		shell = ( abs(x) > abs(y) ) ? abs(x) : abs(y) ;

		// leg -- if you think of the plane of ordered numbers as being inside 
		// a big "X", then the leg is the quadrant in that x.  In other words,
		//	if (x,y) is closest to the positive x axis, then leg is 0
		//	if (x,y) is closest to the positive y axis, then leg is 1
		//	if (x,y) is closest to the negative x axis, then leg is 2
		//	if (x,y) is closest to the negative y axis, then leg is 3
		//
		//	and a "tie" goes to the lower leg number (a tie between legs 3 and 
		//	0 goes to leg 3)
		//
		leg = ( (y <=  x) & (y > -x) ) ? 0 :
			  ( (y >= -x) & (y >  x) ) ? 1 :
			  ( (y >=  x) & (y < -x) ) ? 2 : 3;

		// element is the directed distance from the axis, measured along the 
		// spiral.
		elem = leg == 0 ? y : leg == 1 ? -x : leg == 2 ? -y : x;

		// then n is
		n = (4*shell*shell) + (2*leg-3)*shell + elem;

		return true;
	}

	return false;
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/// Draws a square spiral expanding outwards from x,y
///
/// \param x X co-ordinate for the centre-point for the square spiral
/// \param y Y co-ordinate for the centre-point for the square spiral
/// \param size A measure of how large the finished spiral will be
/// \param rgb Image to write the spiral to
/// \param cmap The colormap to use for writing color on to the pixel
/// \param dim The dimension number for which the spiral is being drawn
///
void sspiral(double startSample, double endSample, Image *rgb, int dim,
			 const QueryDist &query_dist, CHierDisplayInformation *hier_info /*= NULL*/)
{
	okcdata *dataset = g_globals.dataset;
	double min, max;

	// two vector arrays for increased speed
	static int vx[4] = { 0,1,0,-1}, 
			   vy[4] = {-1,0,1, 0};
	
	int drawcount = 0, // number of drawn data points
		count = 0, // number of data points
		dir = 1,   // direction that the spiral will start out in
		len = 1;   // length of the first side to be drawn

	double brushedOpacity = 1.0, unbrushedOpacity = 1.0, opacity = 1.0;
	if (hier_info) {
		samplingFactor = 1.0f;
		if (hier_info->fade_factor>0) 
			unbrushedOpacity = 1-hier_info->fade_factor;
		else if (hier_info->fade_factor<0) 
			brushedOpacity = 1+hier_info->fade_factor;
	}
	RGBt iface_color(interface_color[BG]);

//	int sampled_data_size = 0;
//	for (int c=0; c<query_dist.data_size; ++c) {
//		if (query_dist.randomNums[count] >= start &&
//			query_dist.randomNums[count] <= end)
//		{
//			++sampled_data_size;
//		}
//	}

	int sz = rgb->Width()-1;
	int x = sz/2, y = sz/2;

	int curr_drawcount = 0;
	int tot_drawcount = 0;
	static RGBt rgbColor;

	// while there is data
	while (1)
	{
		for (int j=2; j>0; --j) 
		{
			for (int i=len; i>0; --i)
			{
				++tot_drawcount;
				if (drawcount >= ((sz+1)*(sz+1)))
					return;

				if (curr_drawcount && curr_drawcount < query_dist.sizes[count]) {
					// draw just whites
					rgb->WritePixel(x, y, rgbColor.GetColor());
					// get to next location
					x += vx[dir]; y += vy[dir];
					++curr_drawcount;
					++drawcount;
					continue;
				} else {
					if (curr_drawcount > 0) {
						curr_drawcount = 0;
						++count;
					}
				}

				// if reached at end of data
				if (count >= query_dist.data_size) {
					// draw just whites
					rgb->WritePixel(x, y, interface_color[BG]);
					// get to next location
					x += vx[dir]; y += vy[dir];
					++count;
					curr_drawcount = 0;
					++drawcount;
					continue;
				}

				colormap *cmap;
				int idx = count*(dataset->all_dims+1)+dim;
#ifdef _DEBUG
				if ( (count==0) && (dim==1) ) {
					printf ("count=%d, dim=%d\n", count, dim);
				}
#endif
				double val = query_dist.distances[idx];
				int aaa = 0;

				if (query_dist.randomNums[count] < startSample ||
					query_dist.randomNums[count] > endSample)
				{
					++i; ++count; continue;
				}


				if (dim == 0) {
					if (val < query_dist.brush_level) {
						min = query_dist.dist_min;
						max = query_dist.brush_level;
						cmap = g_globals.pixelBrushedCMap;
						opacity = brushedOpacity;
						aaa = 1;
					} else {
						min = query_dist.brush_level;
						max = query_dist.dist_max;
						cmap = g_globals.pixelUnbrushedCMap;
						opacity = unbrushedOpacity;
						aaa = 2;
					}
				} // if (dim == 0) 
				else {
					Brush &brush = *g_globals.curr_brush;
					// negative values must be below brush center
					if (val <= 0) 
					{
						if (val <= brush.Min(dim-1) - brush.Pos(dim-1)) {
							max = brush.Min(dim-1) - brush.Pos(dim-1);
							min = dataset->all_dim_min[dim-1] - brush.Pos(dim-1);
							cmap = g_globals.pixelUnbrushedCMap;
							opacity = unbrushedOpacity;
							aaa = 3;
						} else {
							max = 0;
							min = brush.Min(dim-1) - brush.Pos(dim-1);
							cmap = g_globals.pixelBrushedCMap;
							opacity = brushedOpacity;
							aaa = 4;
						}
					} // if (val <= 0) 
					else {
						if (val <= brush.Max(dim-1) - brush.Pos(dim-1)) {
							min = 0;
							max = brush.Max(dim-1) - brush.Pos(dim-1);
							cmap = g_globals.pixelBrushedCMap;
							opacity = brushedOpacity;
							aaa = 5;
						} 
						else {
							min = brush.Max(dim-1) - brush.Pos(dim-1);
							max = dataset->all_dim_max[dim-1] - brush.Pos(dim-1);
							cmap = g_globals.pixelUnbrushedCMap;
							opacity = unbrushedOpacity;
							aaa = 6;
						}
					}
				}

				if (val<min)  {
#ifdef _DEBUG
					printf ("val<min, val=%f, min=%f\n", val, min);
#endif
					val = min;
				}
				if (val>max)  {
#ifdef _DEBUG
					printf ("val>max, val=%f, max=%f\n", val, max);
#endif
					val = max;
				}
				assert(val >= min && val <= max);

				double norm_val = (val - min) / (max - min);
				if (val < 0) {
					norm_val = 1.0 - norm_val;
				}
				// color the pixel at the current location
				int index = int((norm_val * (cmap->Len() - 1)) + 0.5);
				rgbColor = (*cmap)[index];
				rgbColor.BlendTo(iface_color, opacity);
				rgb->WritePixel(x, y, rgbColor.GetColor());

				// get to next location
				x += vx[dir]; y += vy[dir];
				// go to next data point
//				++count;
				curr_drawcount = 1;
				++drawcount;
			}
			// should we change direction now?
			dir = (dir + 1) % 4;
		}
		// advance spiral outwards on each half revolution by 1 pixel
		++len;
	}
}
//-----------------------------------------------------------------------------
