////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
//
// Compute and display statistics on data covered by current brush
//

#include <cstdlib>
#include <cmath>

#include <xmdv/stats.h>
#include <xmdv/brush.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/okcdata.h>

static int BP_WIDTH = 200;  // box plot width 
#define MED_DIAM 8      // diameter for filled circle (median)
#define MEAN_DIAM 8     // diameter for open circle (mean)
#define OUT_DIAM 4      // diameter for filled circles (outliers)

// Statistics in double and integer (normalized) forms 
// median, low and upper quartile, lower and upper adjacent values (low/upper
// -/+ 1.5 (upper-lower)), mean, standard deviation 
//

struct stats_tag {
    double median, ql, qu, lav, uav, mean, stdev;
    int imedian, iql, iqu, ilav, iuav, imean, istdev;
    int list_size;
};

static double bmin, bmax;    // brush extents for current dimension

// Compute statistics, determine integer positions within box plot, return
// number of covered points
//
static int comp_double(const void *a, const void *b)
{
    double *fa = (double*) a;
    double *fb = (double*) b;
    if (*fa < *fb) return -1;
    if (*fa > *fb) return 1;
    return 0;
}

// PAUL: stats_org
// Uses BrushMgr's sorted lists of lines to quickly return stats data
int BrushMgr::GetStats(stats_tag & stats, ARR<double> & t_data, int buf_d)
{
    if (!g_globals.dataset) return 0;

	// Initialize summation, size, and brush bounds
	if (ValidLineCount == 0 || Dataset.dims <= 0)
		return 0;

	double total = 0.0;
	
	bmin = g_globals.curr_brush->Min(buf_d);
	bmax = g_globals.curr_brush->Max(buf_d);

	int file_d = Dataset.map_buf_to_file[buf_d];
	int list_size = 0;

	// If the array is empty, no information is stored because
	// all lines have the same value at this dimension.
	if (Index[file_d].Len())
	{
		/*
		Since we already have arrays of sorted line numbers for each
		dimension, it's a good idea to iterate through lines in order
		for this dimension so we do not have to do a sort afterwards.
		*/
		ARR<int> & listOfLineNums = Index[file_d];

		// Get the actual data from indices
		for (int line = 0, j = 0; line < listOfLineNums.Len(); ++line)
		{
			// This is the line's position in the file, rather than
			// in the sorted array.
			int real_line = Index[file_d][line];
			if (!IsLineOutOfBounds(real_line) && IsLineInBrush(real_line, g_globals.curr_brush->num))
			{
				// The line's value at the dimension
				double val = Dataset.data_buf[real_line * Dataset.all_dims + file_d];
				t_data[list_size++] = val;
				total += val;
			}
		}
	}
	else
	{
		// Are ALL lines within the brush range?
		if (Dataset.cg_dim_min[file_d] >= bmin &&
			Dataset.cg_dim_min[file_d] <= bmax)
		{
			list_size = 0;
			// Use all lines
			for (int l = 0; l < Dataset.data_size; ++l)
			{
				if (!IsLineOutOfBounds(l) && IsLineInBrush(l, g_globals.curr_brush->num))
				{
					double val = Dataset.data_buf[l * Dataset.all_dims + file_d];
					t_data[list_size] = val;
					total += val;
					list_size++;
				}
			}
		}
	}
	
	// t_data[j] is ordered because the lines are already inserted in order

	stats.median	= t_data[list_size / 2];
	stats.ql		= t_data[list_size / 4];
	stats.qu		= t_data[3*list_size / 4];

    double range = 1.5 * (stats.qu -stats.ql);
    if ((stats.ql - range) < bmin) 
        stats.lav = bmin;
    else
        stats.lav = stats.ql - range;
    
    if ((stats.qu + range) > bmax)
        stats.uav = bmax;
    else
        stats.uav = stats.qu + range;

	// Compute mean, standard deviation
    stats.mean = total/(double)list_size;
    total = 0.0;
	double tmp;
    
    for (int j = 0;j < list_size;j++)
	{
		tmp = (t_data[j] - stats.mean);
		total += tmp * tmp;
        //total += (t_data[j] - stats.mean) * (t_data[j] - stats.mean);
	}
    
    stats.stdev = sqrt(total/(double)list_size);

    double brange = (bmax - bmin);

    stats.imedian = int(BP_WIDTH * (stats.median - bmin)/ brange);

    // Compute pixel positions within box plot for each statistic
    stats.iql = int(BP_WIDTH * (stats.ql - bmin)/ brange);
    stats.iqu = int(BP_WIDTH * (stats.qu - bmin)/ brange);
    stats.ilav = int(BP_WIDTH * (stats.lav - bmin)/ brange);
    stats.iuav = int(BP_WIDTH * (stats.uav - bmin)/ brange);
    stats.imean = int(BP_WIDTH * (stats.mean - bmin)/ brange);
    stats.istdev = int(BP_WIDTH * (stats.stdev)/ brange);

    stats.list_size = list_size;

    return list_size;
}


int GetStats(stats_tag &stats, ARR<double> &t_data, int i)
{
    if (!g_globals.dataset) return 0;

    ARR<double> buf(g_globals.dataset->dims);

    // Initialize summation, size, and brush bounds
    if (g_globals.dataset->data_size <= 0 || g_globals.dataset->dims <= 0)
        return 0;
        
    double total = 0.0;
    int list_size = 0;
    bmin = g_globals.curr_brush->Min(i);
    bmax = g_globals.curr_brush->Max(i);

    // Fill up list with covered points, updating summation
    int j;
    for (j = 0;j < g_globals.dataset->data_size;j++) {
        g_globals.dataset->GetData(buf, j);
        if (g_globals.curr_brush->IsPointCovered(buf)) {
            t_data[list_size] = buf[i];
            total += buf[i];
            list_size++;
        }
    }

    // Handle empty brush
    if (list_size == 0) return 0;
    
    // Sort list
    qsort(&t_data[0], list_size, sizeof(double), comp_double);

    // Compute median, quartiles, adjacent
    stats.median = t_data[list_size/2];
    stats.ql = t_data[list_size/4];
    stats.qu = t_data[3*list_size/4];

    double range = 1.5 * (stats.qu -stats.ql);
    if ((stats.ql - range) < bmin) 
        stats.lav = bmin;
    else
        stats.lav = stats.ql - range;
    
    if ((stats.qu + range) > bmax)
        stats.uav = bmax;
    else
        stats.uav = stats.qu + range;
    
    // Compute mean, standard deviation
    stats.mean = total/(double)list_size;
    total = 0.0;
    
    for (j = 0;j < list_size;j++)
        total += (t_data[j] - stats.mean) * (t_data[j] - stats.mean);
    
    stats.stdev = sqrt(total/(double)list_size);

    double brange = (bmax - bmin);

    stats.imedian = int(BP_WIDTH * (stats.median - bmin)/ brange);

    // Compute pixel positions within box plot for each statistic
    stats.iql = int(BP_WIDTH * (stats.ql - bmin)/ brange);
    stats.iqu = int(BP_WIDTH * (stats.qu - bmin)/ brange);
    stats.ilav = int(BP_WIDTH * (stats.lav - bmin)/ brange);
    stats.iuav = int(BP_WIDTH * (stats.uav - bmin)/ brange);
    stats.imean = int(BP_WIDTH * (stats.mean - bmin)/ brange);
    stats.istdev = int(BP_WIDTH * (stats.stdev)/ brange);

    stats.list_size = list_size;

    return list_size;
}

// Spacing to use within drawing 
const int gap_small = 3;
const int gap_big = 5;
const int std_shift = 5;
//const int extra = BP_WIDTH/2;

#include <tk.h>

void statsPlotBrushCoverage(Tcl_Interp *interp, Tk_Canvas canvas, const char *canvasname, int dim)
{
    Tk_Window win = Tk_CanvasTkwin(canvas);

    // Allocate space for list of data 
    //
    ARR<double> t_data(g_globals.dataset->data_size);

	Tcl_Eval(interp, "update");
    int W = Tk_ReqWidth(win) - 2;
    int H = Tk_ReqHeight(win) - 2;

    //The following two lines are added by Zaixian because Tk_ReqWidth 
	// and Tk_ReqHeight do not work well in Tcl/Tk 8.4.12
	BP_WIDTH = W - 14;
	//H = 27;
    
    struct stats_tag stats;
	// PAUL: brush_org
    int count = g_globals.dataset->brushMgr->GetStats(stats, t_data, dim);
	//int count = GetStats(stats,t_data,dim);

    char buf[500];

    sprintf(buf," delete all");
    Tcl_VarEval(interp,canvasname,buf,0);
    
    // Draw vertical lines to show brush extents
    //
    char outline[20], fill[20];
    sprintf(outline," -outline #%06lx ",interface_color[STAT1]);
    sprintf(fill," -fill #%06lx ",interface_color[STAT1]);

    sprintf(buf," create line %d %d %d %d", 2 * gap_small, gap_big, 2 * gap_small, H - gap_big);
    Tcl_VarEval(interp,canvasname,buf,fill,0);

    sprintf(buf," create line %d %d %d %d", 2 * gap_small + BP_WIDTH, gap_big, 2 * gap_small + BP_WIDTH, H - gap_big);
    Tcl_VarEval(interp,canvasname,buf,fill,0);

    // Don't drawing anything else
    // if nothing is brushed.
    //
    if (count == 0) return;

    // Draw the median, quartile box, adjacent lines, and outliers in one color 
    //
    sprintf(buf," create rectangle %d %d %d %d", 2 * gap_small + stats.iql, gap_big,
           2 * gap_small + stats.iqu, H - gap_big);
    Tcl_VarEval(interp,canvasname,buf,outline,0);

    sprintf(buf," create arc %d %d %d %d -start 0 -extent 359 -style pieslice ", 
        2 * gap_small + stats.imedian - MED_DIAM/2, 
        H/2 - MED_DIAM/2,
        2 * gap_small + stats.imedian + MED_DIAM/2 +1, 
        H/2 + MED_DIAM/2);
    Tcl_VarEval(interp,canvasname,buf,outline,fill,0);

    // adjacent lines are dashed, sort of side-ways T's (should be square brackets,
    // to be "official" Tukey box plots)
    //

    sprintf(buf," create line %d %d %d %d ",
        2 * gap_small + stats.ilav,
        gap_big,  
        2 * gap_small + stats.ilav,
        H - gap_big);
    Tcl_VarEval(interp,canvasname,buf,fill,0);

    sprintf(buf," create line %d %d %d %d ",
       2 * gap_small + stats.iuav, 
       gap_big, 
       2 * gap_small + stats.iuav,
       H - gap_big); 
    Tcl_VarEval(interp,canvasname,buf,fill,0);

    sprintf(buf," create line %d %d %d %d ",
       2 * gap_small + stats.iql,
       H/2, 
       2 * gap_small + stats.ilav,
       H/2); 
    Tcl_VarEval(interp,canvasname,buf,fill,0);

    sprintf(buf," create line %d %d %d %d ",
       2 * gap_small + stats.iqu, 
       H/2, 
       2 * gap_small + stats.iuav,  
       H/2); 
    Tcl_VarEval(interp,canvasname,buf,fill,0);

	// Figure out which points are outliers (outside adjacent points) and plot
/*
	for (int j = 0;j < count;j++) {
        if (t_data[j] > stats.uav || t_data[j] < stats.lav) {
            int outx = int(BP_WIDTH * (t_data[j] - bmin) / (bmax - bmin));

            sprintf(buf," create arc %d %d %d %d -start 0 -extent 359 -style pieslice ", 
                2 * gap_small + outx - OUT_DIAM/2,
                H/2 - OUT_DIAM/2,
                2 * gap_small + outx + OUT_DIAM/2, 
                H/2 + OUT_DIAM/2);
            Tcl_VarEval(interp,canvasname,buf,outline,fill,0);
        }
    }
*/
    // Figure out which points are outliers (outside adjacent points) and plot
    //
	const double cfactor = BP_WIDTH / (bmax - bmin);
	// PAUL: brush_org
	// Break the job into 2 parts, to speed things up quite a bit
	for (int j = 0;j < count && t_data[j] < stats.lav;j++)
	{
		int outx = int(cfactor * (t_data[j] - bmin));

		sprintf(buf," create arc %d %d %d %d -start 0 -extent 359 -style pieslice ", 
			2 * gap_small + outx - OUT_DIAM/2,
            H/2 - OUT_DIAM/2,
			  2 * gap_small + outx + OUT_DIAM/2, 
          H/2 + OUT_DIAM/2);
      Tcl_VarEval(interp,canvasname,buf,outline,fill,0);
	}
	for (int j = count-1;j >= 0 && t_data[j] > stats.uav; j--)
	{
		int outx = int(cfactor * (t_data[j] - bmin));

        sprintf(buf," create arc %d %d %d %d -start 0 -extent 359 -style pieslice ", 
            2 * gap_small + outx - OUT_DIAM/2,
            H/2 - OUT_DIAM/2,
            2 * gap_small + outx + OUT_DIAM/2, 
            H/2 + OUT_DIAM/2);
        Tcl_VarEval(interp,canvasname,buf,outline,fill,0);
	}

	// Change color and draw mean and standard deviation 
    //
    sprintf(outline," -outline #%06lx ",interface_color[STAT2]);
    sprintf(fill," -fill #%06lx ",interface_color[STAT2]);

    sprintf(buf," create arc %d %d %d %d -start 0 -extent 359 -style arc ",
        2 * gap_small + stats.imean - MEAN_DIAM/2, 
        H/2 - MEAN_DIAM/2, 
        2 * gap_small + stats.imean + MEAN_DIAM/2, 
        H/2 + MEAN_DIAM/2);
    Tcl_VarEval(interp,canvasname,buf,outline,fill,0);
    
    // Standard deviation is indicated by angle brackets
    //
    int stdpos = stats.imean - stats.istdev;
    sprintf(buf," create line %d %d %d %d ",
       2 * gap_small + stdpos, 
       H/2, 
       2 * gap_small + stdpos + std_shift,
       H -gap_big);
    Tcl_VarEval(interp,canvasname,buf,fill,0);

    sprintf(buf," create line %d %d %d %d ",
       2 * gap_small + stdpos,
       H/2, 
       2 * gap_small + stdpos + std_shift,
       gap_big);
    Tcl_VarEval(interp,canvasname,buf,fill,0);
    
    stdpos = stats.imean + stats.istdev;

    sprintf(buf," create line %d %d %d %d ",
       2 * gap_small + stdpos,
       H/2, 
       2 * gap_small + stdpos - std_shift,
       H - gap_big);
    Tcl_VarEval(interp,canvasname,buf,fill,0);

    sprintf(buf," create line %d %d %d %d ",
       2 * gap_small + stdpos,
       H/2, 
       2 * gap_small + stdpos - std_shift,
       gap_big);
    Tcl_VarEval(interp,canvasname,buf,fill,0);
}
