#include <CacheManager.h>
#include <Cache.h>
#include <ReplacementPolicy.h>
#include <Brush.h>
#include <LRUReplacement.h>
#include <list>
#include <string>
#include <BrushVisitor.h>
#include <QueryFactory.h>
#include <Loader.h>
#include <RegionFactory.h>
#include <Record.h>
#include <IReplacementVisitor.h>
#include <ICacheEntry.h>
#include <CacheIterator.h>
#include <Mapper.h>
#include <cache/PerformanceStatistics.h>
#include <Timer/Timer.h>
#include <cache/IFragThresholdVisitor.h>

using namespace cache;
using namespace std;


CacheManager::CacheManager(Loader* loader,
						   Cache* cache,
						   ReplacementPolicy *replacementPolicy,
						   AbstractQueryFactory *queryFactory,
						   unsigned int fragmentationThresold):
cacheLogger("CacheManager","Testing CacheManager"),
debugStream("debugFile"), //Used for debugging purposes
callNumber(0),
fragmentationThreshold(fragmentationThresold)
{
	this->timer = new ::Timer();

	this->queryFactory = queryFactory;

	this->cache = cache;

	this->replacementPolicy = replacementPolicy;

	this->loader = loader;

	this->cacheEntryQueue = new std::queue<ICacheEntry*>();		

	//Registering the replacement policy to hear the servedItem Event
	this->addServedItemListener(replacementPolicy);
	//Note this should be in the CacheBuilder

	brushVisitor = static_cast<BrushVisitor*> (replacementPolicy);				
	this->numberOfTuplesLoaded = 0;

	mutex = new ZThread::Mutex();
}
/*Instantiates the cache*/		
/*Instantiates the replacement policy*/
/*Instantiates the brush*/
/*Registers the replacement policy with the brush Visitor*/
/*Instantiates the Loader*/				


CacheIterator* CacheManager::load(Brush &brush,
								  IReplacementVisitor* replacementVisitor,
								  IFragThresholdVisitor* fragThresholdVisitor,
								  AbstractRegionFactory* regionFactory,
								  PerformanceStatistics& perStatistics,
								  QueryModifier& queryModifier)
{
	mutex->acquire();

	//The number of tuples loaded in the pervious call.
	perStatistics.brushSize = numberOfTuplesLoaded;

	callNumber = callNumber + 1;

	#ifdef _DEBUG
	LOG_GMSG(cacheLogger,"In Load CallNumber "<<callNumber);
	#endif
	if(callNumber == 154)
		std::cout<<"break";

	CacheIterator* cacheIterator = NULL;
	//The mutex will be released only in stop load.
	//You want to make sure that there is only one thread
	//That completes the load and all other requests for load
	//have to wait unti that completes.		
	this->replacementVisitor = replacementVisitor;
	this->currentBrush = brush;		
	this->regionFactory = regionFactory;
	this->brushVisitor->brushModified(brush);		

	this->timer->startTimer();

	std::list<ICacheEntry*> &listCacheEntry =
		this->cache->searchNode(currentBrush);	

	this->searchTime = this->timer->endTimer();
	perStatistics.searchLatency  = this->searchTime;

	std::vector<RemainderRegion*>& vecRemainderRegions = 
		regionFactory->makeRegions(listCacheEntry,currentBrush);

	perStatistics.totalFragments = vecRemainderRegions.size();	


	if(vecRemainderRegions.size() > 0)
	{
		//Calling the brushVisitor method to notify the replacement policy
		this->replacementPolicy->ifRemainder(true);
		//Only for debugging purposes
		this->replacementPolicy->checkForError();


		if(vecRemainderRegions.size() > fragmentationThreshold) //Limit to the max number of remainder regions
		{
			if(fragThresholdVisitor->exceededFragThreshold())
			{
			perStatistics.numberOfHits = listCacheEntry.size();
			//in case the cache is completely defragmented we need to take severe steps
			//to see to it that the query does not exceed the max number of open cursors
			this->replacementPolicy->removeEntries(listCacheEntry);
			//Load the complete brush that is we completely by pass the contents of the cache
			currentQuery = queryFactory->makeQuery(currentBrush);
			currentQuery = queryModifier.modifyQuery(currentQuery);
			//Clear the cache entry queue			
			#ifdef _DEBUG
			LOG_GMSG(cacheLogger,"!!!!!!!!!!!!Cleared the cache!!!!!!!!!!!!!");
			LOG_GMSG(cacheLogger,"The query is :"<<currentQuery.c_str());
			#endif
			debugStream<<callNumber<<"  "<<currentQuery<<std::endl<<std::endl;
			this->recordIterator = loader->load(currentQuery);								
			}
		}
		else
		{						
			perStatistics.numberOfHits = listCacheEntry.size();
			currentQuery = queryFactory->makeQuery(vecRemainderRegions);		
			currentQuery = queryModifier.modifyQuery(currentQuery);
			#ifdef _DEBUG
			LOG_GMSG(cacheLogger,"The query is :"<<currentQuery.c_str());
			#endif
			debugStream<<callNumber<<"  "<<currentQuery<<std::endl<<std::endl;
			this->recordIterator = loader->load(currentQuery);							
			//Now pushing all the hit entries into the cacheEntryQueue
			this->handleHits(listCacheEntry);
		} // if(vecRemainderRegions.size() > 0)
	}
	else
	{
		perStatistics.numberOfHits = listCacheEntry.size();
		//Now pushing all the hit entries into the cacheEntryQueue
		this->handleHits(listCacheEntry);
		this->recordIterator = NULL;
	}

	#ifdef _DEBUG
	LOG_GMSG(cacheLogger,"After Loading the query");				
	#endif
	//searches through the cache
	//makes the remainder regions
	//fills the queue with the hits in the cache
	//instantiates the iterator.				
	cacheIterator = new CacheIterator(this,
									this->cacheEntryQueue,
									&this->servedItemVisitorImpl);		

	return(cacheIterator);
}


CacheIterator* CacheManager::load(Brush &brush,
								  IReplacementVisitor* replacementVisitor,
								  IFragThresholdVisitor* fragThresholdVisitor,
								  AbstractRegionFactory* regionFactory,
								  PerformanceStatistics& perStatistics)//Returns the number of hits in the cache
{
	QueryModifier qmf (0);
	return(this->load(brush,
					replacementVisitor,
					fragThresholdVisitor,
					regionFactory,
					perStatistics,
					qmf));
}

/*
* This function updates the state of the replacement policy to notify it of the hits
* in the cache. Also it inserts the cache hits into the cache entry queue. This will work 
* for the cache.
*/
void CacheManager :: handleHits(std::list<ICacheEntry*>& listCacheEntry)
{
	std::list<ICacheEntry*>::const_iterator it;

	for(it = listCacheEntry.begin();it != listCacheEntry.end();++it)
	{
		this->replacementPolicy->hitCacheEntry(**it);
		this->cacheEntryQueue->push(*it);
	}	
}

void CacheManager:: loadLeftOver()
{
	//try{	
	#ifdef _DEBUG
	LOG_GMSG(cacheLogger,"In Load Left Over");
	#endif
	if(this->recordIterator != NULL)
	{		
		while(!recordIterator->isDone())
		{		
			ICacheEntry* cacheEntry = this->replacementPolicy->peek();

			if(!this->replacementVisitor->chosenItem(*cacheEntry))
				break; //If false then discontinue and don't replace the item
			
			cacheEntry = 
				this->replacementPolicy->getVictimItem();			

			if(cacheEntry == NULL)//No space in the cache used for incremental loading
				break;

			Record *record = recordIterator->get_next();
			Mapper::getInstance()->initialize(*cacheEntry,*record);			
			cache->addCacheEntry(*cacheEntry);		
			this->cacheEntryQueue->push(cacheEntry);
			//Don't need the record any more
			delete(record);						 
		} // while(!recordIterator->isDone())
	} // if(this->recordIterator != NULL)		

	/*}catch(...)
	{
		#ifdef _DEBUG
		LOG_GMSG(cacheLogger,"Got the Exception");
		LOG_GMSG(cacheLogger,"The current query is :"<<currentQuery);
		#endif
		debugStream.close();
		#ifdef _DEBUG
		LOG_GMSG(cacheLogger,"Throwing Exception");
		#endif
		throw(1);		
	}*/
}

/*
* This returns the time taken to search in the cache
* in msec
*	@return unsigned long
*/
unsigned long CacheManager::getSearchTime()
{
	return(this->searchTime);
}
/**
* Deletes the record Iteration process. Should work
* even if all the objects are not read from the 
* iterator
*/
void CacheManager:: endLoad(unsigned long count)
{	
	//Because stop load can spawn a new thread that calls load					

	//End of loading	
	if(recordIterator != NULL)
		delete(recordIterator);
	recordIterator = NULL;
	//This is used to inform the loadListener that the loading has
	//ended						
	this->replacementVisitor = NULL;
	this->regionFactory = NULL;			
	this->numberOfTuplesLoaded = count;
	mutex->release();	
}


//load tuples regions until the cache is full.
//For each new entry
//a) remove one entry from the cache
//b) Add a new entry into the cache
//c) Store the pointer in the queue	


/**
* Add the served Listener
* \param *servedItemVisitor 
*/
void CacheManager:: addServedItemListener(
	ServedItemListener *servedItemListener)
{
	servedItemVisitorImpl.
		addServedItemListener(servedItemListener);
}

CacheManager:: ~CacheManager()
{		
	delete(timer);
	delete(cache);
	delete(cacheEntryQueue);
	delete(replacementPolicy);	
	delete(mutex);
	//Deleting all the singletons	
	delete(Mapper::getInstance());				
}
