#include <cache/QueryFactory.h>
#include <stdio.h>
#include <string>
#include <sstream>
#include <vector>
#include <cache/RemainderRegion.h>
#include <cache/Brush.h>
#include <zthread/Thread.h>
#include <iostream>
using namespace cache;

QueryGeomFactory* QueryGeomFactory::qf = NULL;
QuerySimpleFactory* QuerySimpleFactory::qsf = NULL;

AbstractQueryFactory::AbstractQueryFactory(Schema *schema,std::string tableName)
				:outputFile("query.txt")
{
	this->schema = schema;
	this->tableName = tableName;
	this->selectClause = this->makeSelectClause();
}


std::string AbstractQueryFactory::makeSelectClause()
{
	std::ostringstream stringStream; 
	stringStream<<"select /*+NO_CACHE*/ "
		<<this->schema->getSelectString()
		<<" from "
		<<this->tableName
		<<" where ";		
	return(stringStream.str());

}


/*
*	QueryGeomFactory is a singleton
* It is used to make the query to be sent to the 
* data base, It uses the RemainderRegions and the 
* table Name to form the query. Note this class
* can be used only for the structure based brush.
*/
QueryGeomFactory::QueryGeomFactory(Schema *schema,std::string tableName)
				  :AbstractQueryFactory(schema,tableName)
{
	

}

QueryGeomFactory* QueryGeomFactory :: makeQueryFactory(Schema *schema,std::string tableName)
{
	if(qf == NULL)
		qf = new QueryGeomFactory(schema,tableName);
	return(qf);
}

/**
* Constructs the geometry predicate to use the R Tree index at the 
* backend.
* \param &r 
* \return 
*/
std::string QueryGeomFactory::getGeometryPredicate(double leftExtent,
												   double rightExtent,
												   double levelOfDetail)
{
	std::ostringstream stringStream;
	stringStream<<"SDO_FILTER(shape,MDSYS.SDO_GEOMETRY("
		<<" 2002," 
		<<" NULL,"
		<<" NULL,"
		<<" MDSYS.SDO_ELEM_INFO_ARRAY(1,2,1)," 
		<<" MDSYS.SDO_ORDINATE_ARRAY("
		<<leftExtent<<","
		<<levelOfDetail<<","
		<<rightExtent<<","
		<<levelOfDetail
		<<")),"
		<<" 'querytype = window' "
		<<")"
		<<"="
		<<" 'TRUE'";

	return(stringStream.str());
}

/**
* Makes the query
* \param &vecRemRegions 
* \param tableName 
* \return 
*/
std::string QueryGeomFactory:: makeQuery(std::vector<RemainderRegion*>& vecRemRegions)
{
	if(vecRemRegions.size() == 0)
		return "";

	std::ostringstream stringStream;
	//Generates the 1st predicate for the 1st remainder region
	stringStream<<this->selectClause;		

	stringStream<<this->getGeometryPredicate(vecRemRegions[0]->getLeftExtent(),
											vecRemRegions[0]->getRightExtent(),
											vecRemRegions[0]->getLevelOfDetail())
				<<" and "				
				<<vecRemRegions[0]->getPredicate()
				<<" and "
				<<" L <= "
				<<vecRemRegions[0]->getLevelOfDetail()
				<<" and "
				<<" PL > "
				<<vecRemRegions[0]->getLevelOfDetail();

	//Outputting the query to the file.
	/*outputFile<<this->selectClause<<this->getGeometryPredicate(vecRemRegions[0]->getLeftExtent(),
		vecRemRegions[0]->getRightExtent(),
		vecRemRegions[0]->getLevelOfDetail())
		<<" and "				
		<<vecRemRegions[0]->getPredicate()
		<<" and "
		<<" L <= "
		<<vecRemRegions[0]->getLevelOfDetail()
		<<" and "
		<<" PL > "
		<<vecRemRegions[0]->getLevelOfDetail()
		<<std::endl;*/

	
	//Generates the predicate for the rest of the remainder regions.
	for(unsigned int i=1;i<vecRemRegions.size();i++)		
	{
		stringStream<<" union ";

		stringStream<<selectClause<<this->getGeometryPredicate(vecRemRegions[i]->getLeftExtent(),
									 vecRemRegions[i]->getRightExtent(),
									 vecRemRegions[i]->getLevelOfDetail())
		<<" and "
		<<vecRemRegions[i]->getPredicate()
		<<" and "
		<<"L <= "<<vecRemRegions[i]->getLevelOfDetail()
		<<" and "		
		<<" PL > "<<vecRemRegions[i]->getLevelOfDetail();
	
		//Outputing the query
		/*outputFile<<selectClause<<this->getGeometryPredicate(vecRemRegions[i]->getLeftExtent(),
			vecRemRegions[i]->getRightExtent(),
			vecRemRegions[i]->getLevelOfDetail())
			<<" and "
			<<vecRemRegions[i]->getPredicate()
			<<" and "
			<<"L <= "<<vecRemRegions[i]->getLevelOfDetail()
			<<" and "		
			<<" PL > "<<vecRemRegions[i]->getLevelOfDetail()
			<<std::endl;*/
	}
		

	//Outputting the complete query
	//outputFile<<stringStream.str()<<std::endl<<"#####"<<std::endl;

	return(stringStream.str());		
}

/**
* Makes the query from the brush
* \param brush 
* \return 
*/
std::string QueryGeomFactory :: makeQuery(cache::Brush& brush)
{
	std::ostringstream stringStream;

	stringStream<<this->selectClause
				<<this->getGeometryPredicate(brush.getLeftExtent(),
											  brush.getRightExtent(),
											  brush.getLevelOfDetail())
				<<" and "
				<<brush.getPredicate()
				<<" and "
				<<" L <= "<<brush.getLevelOfDetail()
				<<" and "
				<<" PL > "<<brush.getLevelOfDetail();

	return(stringStream.str());
}


/**
 * Returns the reference of type QueryGeomFactory
 * \return 
 */
QueryGeomFactory*& QueryGeomFactory::getReference()
{
	return(qf);
}
/**
* Destorys the query factory
* \return 
*/
QueryGeomFactory::~QueryGeomFactory()
{}


QuerySimpleFactory::QuerySimpleFactory(Schema *schema,std::string tableName):
					AbstractQueryFactory(schema,tableName)
{

}

/**
* This is used to make the query factory
* \param schema 
* \param tableName 
* \return 
*/
QuerySimpleFactory* QuerySimpleFactory::makeQueryFactory(Schema* schema,
											std::string tableName)
{
	if(qsf == NULL)
		qsf = new QuerySimpleFactory(schema,tableName);
	return(qsf);
}

/**
 * Returns the reference to the pointer of type query simple factory
 * \return 
 */
QuerySimpleFactory*& QuerySimpleFactory::getReference()
{
	return(qsf);
}
/**
* Makes the query from the brush
* \param brush 
* \return 
*/
std::string QuerySimpleFactory :: makeQuery(cache::Brush& brush)
{	

	std::ostringstream stringStream;
	//Just simulating a bigger database. Actually useful only in case we have 
	//a small database.
	ZThread::Thread::sleep(200);

	stringStream<<this->selectClause;

	stringStream<<" ( "				
		<<brush.getPredicate()				
		<<" ) "
		<<" and L <= "<<brush.getLevelOfDetail()
		<<" and PL > "<<brush.getLevelOfDetail();
		
	return(stringStream.str());
}


/**
* Makes the query
* \param &vecRemRegions 
* \param tableName 
* \return 
*/
std::string QuerySimpleFactory::makeQuery(std::vector<RemainderRegion*>& vecRemRegions)
{
	if(vecRemRegions.size() == 0)
		return "";

	std::ostringstream stringStream;

	//Just simulating a bigger database. Actually useful only in case we have 
	//a small database.
	ZThread::Thread::sleep(200);

	stringStream<<this->selectClause;

	stringStream<<" ( ";

	RemainderRegion* remainderRegion = vecRemRegions[0];
	
	std::ostringstream queryStream;
	//Writes the query to the output file
	queryStream<<this->selectClause<<" ( "<<remainderRegion->getPredicate()<<" ) "
			<<" and (L <= "<<vecRemRegions[0]->getLevelOfDetail()
			<<" and PL > "<<vecRemRegions[0]->getLevelOfDetail()<<")"
			<<std::endl;

	std::string queryString = queryStream.str();

	/*outputFile<<queryString;*/

	stringStream<<" ( "				
		<<remainderRegion->getPredicate()				
		<<" ) ";
	
	for(unsigned int i=1;i<vecRemRegions.size();i++)		
	{
		stringStream<<" or "
		<<" ( "		
		<<vecRemRegions[i]->getPredicate()
		<<" ) ";

		//Past a query for each remainder predicate
		/*outputFile<<this->selectClause<<" ( "		
			<<vecRemRegions[i]->getPredicate()
			<<" ) "
			<<" and (L <= "<<vecRemRegions[0]->getLevelOfDetail()
			<<" and PL > "<<vecRemRegions[0]->getLevelOfDetail()<<")"
			<<std::endl;*/
	}

	stringStream<<" ) ";

	
	
	stringStream<<" and (L <= "<<vecRemRegions[0]->getLevelOfDetail()
				<<" and PL > "<<vecRemRegions[0]->getLevelOfDetail()<<")";
	
	//Paste the final query.
	//outputFile<<stringStream.str()<<std::endl<<"#####"<<std::endl;

	return(stringStream.str());		
}

/**
* Destorys the query factory
* \return 
*/
QuerySimpleFactory::~QuerySimpleFactory()
{}



