#include <cache/prefetcher/Prefetcher.h>
#include <cache/CacheEntry.h>
#include <cache/CacheIterator.h>
#include <cache/CacheManager.h>
#include <zthread/Thread.h>
#include <zthread/Semaphore.h>
#include <cache/Brush.h>
#include <db_GLOBAL.h> //Contains the definitions for all the macros
#include <strategyTable.h> //Defines the methods for the strategy table
#include <db_DISCOVER1.h>
#include <cache/BrushStore.h>
#include <cache/RegionFactory.h>
#include <cache/PrefetcherHelper.h>
#include <PreEmptException.h>
#include <PrefetchVisitor.h>
#include <log/log.h>
#include <cache/LoadRequester.h>
#include <cache/ICacheEntry.h>
#include <cache/PerformanceStatistics.h>
#include <cache/prefetcher/PrefetchController.h>
#include <cache/QueryFactory.h>

using namespace cache;

Prefetcher::Prefetcher(const unsigned long sizeOfCache,
					   unsigned int numberOfLevels,//Used temporarily
					   BrushStore* brushStore,		
					   int AdaptivePrefetching,
					   PrefetchVisitor *prefetchVisitor,
					   unsigned int PrefetchStrategy, //Initial Prefetch Strategy 1 for random
					   unsigned int NO_P_ENTRIES//This is also temporary		
					   ):prefetchLogger("prefetcher","Logs the statements in the prefetcher")
{
	this->semPrefetchPermit = NULL;
	this->semLoadPermit = NULL;
	this->DO_ADAPTIVE_PREFETCHING = AdaptivePrefetching;
	this->preempt = false;
	this->sizeOfCache = sizeOfCache;		
	this->flag = false;		
	this->cacheIterator = NULL;
	this->prefetchVisitor = prefetchVisitor;
	this->Prefetch_Strategy = PrefetchStrategy;				
	prefetchLogger.enableMsgGroup();		

	/*Temporary*************************************************/
	this->numberOfLevels = numberOfLevels;//Temporary
	this->NO_P_ENTRIES = NO_P_ENTRIES;//Temporary
	this->brushType = 0;
	this->brushStore = brushStore;

	//Making the xmdvFocus 2 dimensional array
	//This should not be hard coded but this was the way it was
	xmdvFocus= new double*[5];
	for(int i=0;i<5;i++)
		xmdvFocus[i] = new double[3];

	for(int i=0;i<5;i++)
	{
		for(int j=0;j<3;j++)
		{
			xmdvFocus[i][j] = 0.0;
		}
	} // for(int i=0;i<5;i++)
	this->DISC_FOC = 0;
	/*Temporary*************************************************/
}

/**
* Destroys the Prefetcher
* \return 
*/
Prefetcher::~Prefetcher()
{
	for(int i=0;i<5;i++)
		delete[] (xmdvFocus[i]);
	delete[] xmdvFocus;

	for(unsigned int i=0;i<vecPrefetchBrushes.size();i++)
		delete(vecPrefetchBrushes[i]);
}

/**
* Starts the Prefetcher
*/
void Prefetcher::run()
{	
	try{
		while(true)
		{							
			this->preempt = false;			
			//Consuming the permit given to us by the cachecontroller.
			this->semPrefetchPermit->acquire();
	#ifdef _DEBUG
			LOG_GMSG(prefetchLogger,"Prefetcher Now Running");
	#endif
			this->do_prefetching();	
			this->cleanup();
			//Releases the Semaphore
			this->semLoadPermit->release();				
		} // while(1)
	}catch(ZThread::Interrupted_Exception ie)
	{				
	#ifdef _DEBUG
		LOG_GMSG(prefetchLogger,"Prefetcher Got Interrupted Exception");
	#endif
		::std::cout<<"Prefetcher Terminating"<<::std::endl;			
	}
	catch(...)			
	{
		::std::cout<<"Got exception in run : error"<<::std::endl;
	}
}

/**
* Used by the prefetch Controller to preempt the prefetcher
* \param preEmpt 
*/
void Prefetcher::setPreEmpt(bool preEmpt)
{
	this->preempt = preEmpt;
}

/*
*	Returns the pointer to the prefetch brushes.
*/
::std::vector<Brush*>& Prefetcher::getPrefetchBrushes()
{
	return(this->vecPrefetchBrushes);
}

/**
*
* \Decides whether the prefetcher wants to continue or not 
* Note it depends upon the current replacement candidate whether
* the prefetcher want to continue or not.
* \return bool
*/	
bool Prefetcher::ifContinue(ICacheEntry& cacheEntry)
{		
	if(preempt == true)
		return(false);

	Brush* currentBrush = brushStore->getBrush(0);

	Tools::Geometry::IShape* iShape = cacheEntry.getIShape();

	if(iShape != NULL)//If the node is not free
	{
		if(currentBrush->inBrush(*iShape))
			return(false);

		for(unsigned int i=0;i<vecPrefetchBrushes.size();i++)
		{
			if(vecPrefetchBrushes[i]->inBrush(*iShape))
				return(false);
		} // for(unsigned int i=0;i<vecPrefetchBrushes.size()&&result;i++)					
	} // if(iShape != NULL)//If the node is not free
	return(true);
} 

/**
* Returns the number of nodes prefetched by the prefetcher
* \return unsigned int
*/
unsigned int Prefetcher::getNumberOfNodesPrefetched()
{
	return(this->numberOfNodesPrefetched);
}


//Methods to maintain a consistent state
void Prefetcher::cleanup()
{
	//LOG_GMSG(prefetchLogger,"Prefetcher cleaning up");
	//calling the Visitor	
	if(prefetchVisitor != NULL)
		this->prefetchVisitor->prefetcherShuttingDown(this);
	
	this->numberOfNodesPrefetched = 0;

	//Clearing prefetch Brushes note there can be more than one
	//prefetch brushes in a single run of the prefetcher		
	for(unsigned int i=0;i<vecPrefetchBrushes.size();i++)
	{
		delete(vecPrefetchBrushes[i]);
	}
	vecPrefetchBrushes.clear();			
	#ifdef _DEBUG
	LOG_GMSG(prefetchLogger,"Prefetch Thread cleaned up: giving up CPU");
	#endif
}


//Methods imported from old prefetcher
/*************************************************************************
init_dir

If prefetch strategy is random, make discover paths in all
the directions=0.25.
For the remaining strategies make it 0.1
*************************************************************************/

void Prefetcher::init_dir(double *vect)
{
	int index;
	double value ;
	//When Prefetch strategy is 1 that means random prefetch.
	//If other 2 strategies the value = 
	if (DO_ADAPTIVE_PREFETCHING == 0)
		value = Prefetch_Strategy==1 ? .25 : (1.0-DISC_DIR)/3.0;
	else
		value = (strategyTable_GetSelectedStrategy() == RANDOM_PREFETCH)? .25 : (1.0-DISC_DIR)/3.0;
	// Adaptive Prefetching : APrefetchStrategy = Prefetch_Strategy

	for (index=0; index<4; index++) vect[index] = value;
}


/*************************************************************************
adjust_prefetch_values

Depending on the direction in which to prefetch,
assign the focus values from the focus table.
*************************************************************************/

void Prefetcher::adjust_prefetch_values(int direction, double *up, double *dn,
										double *le, double *ri)
{
	Brush* currentBrush = this->brushStore->getBrush(0);
	double	x = currentBrush->getLeftExtent(),
		y = currentBrush->getRightExtent(),
		l = currentBrush->getLevelOfDetail();
	int index;


	if (direction==0)
		for (index=0; index<DISC_FOC; index++)
			if (xmdvFocus[index][2] > l)
			{
				*up = xmdvFocus[index][2];
				goto exit;
			}

			if (direction==1)
				for (index=0; index<DISC_FOC; index++)
					if (xmdvFocus[index][2] < l)
					{
						*dn = xmdvFocus[index][2];
						goto exit;
					}

					if (direction==2)
						for (index=0; index<DISC_FOC; index++)
							if (xmdvFocus[index][0] < x)
							{
								*le = xmdvFocus[index][0];
								goto exit;
							}

							if (direction==3)
								for (index=0; index<DISC_FOC; index++)
									if (xmdvFocus[index][1] > y)
									{
										*ri = xmdvFocus[index][1];
										goto exit;
									}

exit:
									::std::cout<<"";	
}

/*************************************************************************
calculateStatisticsForFitnessFunction

This function counts the number of objects predicted by the
given prefetching scheme

*************************************************************************/
/*
*	This function will change completely because we do 
* have a different cacheStrcuture and we cannot do something like this
* This same task is acheived in the following manner
* a) Summing over the number of hits each time we call the load in the prefetch
* b) Summing over the number of new items loaded this can be done in ifContinue()
*/
void Prefetcher::calculateStatisticsForFitnessFunction(double x, double y, double level)
{}

/*************************************************************************
do_prefetching

Call init_dir()
For prefetch strategy=2 or 3, set discover paths to 0.6
if there are more regions to be discovered in that direction.

start loop

calculate the probability in the direction in which to move.
if prefetch strategy is 'focus', then adjust the prefetch
values by calling adjust_prefetch_values()

determine what data to be loaded from the database into memory
in each directions by calling preparePrefetching() 
separately for each cirections.

end loop

(ger) revised to use preparePrefetching instead of prepare() to work with
or without incremental loading
*************************************************************************/

void Prefetcher::do_prefetching()
{
	try{		
		//LOG_GMSG(prefetchLogger,"In Do Prefetching ");	
		//Both these brushes are user brushes
		Brush* oldBrush = brushStore->getBrush(1);
		Brush* currentBrush = brushStore->getBrush(0);		

		if(oldBrush == NULL)
			return; //The prefetcher starts prefetching only after 2 user brushes

		double 
			leftExtent  = currentBrush->getLeftExtent(),
			rightExtent = currentBrush->getRightExtent(),
			levelExtent = currentBrush->getLevelOfDetail();

		double	up,
			down,
			left,
			right;	    /* left/right/up/down prefetch values */

		double	dir[4];		/* up-0, down-1, left-2, right-3 */
		//Note the 4 elements of the array store  the value for 
		//each of the 4 directions the brush can move 1) up 
		//2) down, 3) left 4) 

		double	i,
			delta;

		int	direction, p;		/* direction & probability */

		/*********************************temporary**********************/
		delta = 1.0/numberOfLevels;
		/*********************************temporary**********************/			
		init_dir(dir);		
		if ( ((DO_ADAPTIVE_PREFETCHING == 0) && 
			(Prefetch_Strategy == DIRECTION_PREFETCH || Prefetch_Strategy == FOCUS_PREFETCH))
			||
			((DO_ADAPTIVE_PREFETCHING == 1) &&
			(strategyTable_GetSelectedStrategy() == DIRECTION_PREFETCH || 
			strategyTable_GetSelectedStrategy() == FOCUS_PREFETCH)))
		{
			if (levelExtent > oldBrush->getLevelOfDetail()) 
			{
				dir[0] = DISC_DIR;
				if ( levelExtent+delta > 1 )
					dir[0] = 0 ;
			}

			if (levelExtent < oldBrush->getLevelOfDetail())
			{
				dir[1] = DISC_DIR;
				if ( levelExtent-delta < 0 )
					dir[1] = 0 ;
			}

			if (leftExtent < oldBrush->getLeftExtent())
			{
				dir[2] = DISC_DIR;
				if ( leftExtent <= 0 )
					dir[2] = 0 ;
			}

			if (rightExtent > oldBrush->getRightExtent())
			{
				dir[3] = DISC_DIR;
				if ( rightExtent >= 1 )
					dir[3] = 0 ;
			}

		}

		//Returns the index of the max value in the array, note the index determines 
		//the direction. The max value simply associates a weight with the direction.
		direction = prefetchHelper.undeterministic_max(dir, 4);
		
		//determines the direction value associated with the direction.
		p = (int)((NO_P_ENTRIES-1)*dir[direction]);
		if (!p)
			goto exit ;

		up = 1.0;
		down = 0.0;
		left = 0.0;
		right = 1.0;		


		if (  (DO_ADAPTIVE_PREFETCHING == 0 && Prefetch_Strategy == FOCUS_PREFETCH)
			||
			(DO_ADAPTIVE_PREFETCHING == 1 && strategyTable_GetSelectedStrategy() == FOCUS_PREFETCH) 	)  
			adjust_prefetch_values(direction, &up, &down, &left, &right);

		#ifdef _DEBUG
		LOG_GMSG(prefetchLogger,"Prefetch Direction 0=up,1=down,2=left,3=right :"<<direction);		
		#endif

		if (direction==0)//Indicates prefetch towards the top of the tree.
		{
			regionFactory = AscRegionFactory::getInstance();				
			int count = 0;
			for (i=levelExtent+delta; (i+delta)<=up; i+=delta,count++)
			{
				//ger							
				QueryModifier qf(1);
				this->loadBrush(leftExtent,rightExtent,i, regionFactory, qf);								
				if ( (i > 1) || (i < 0) )	goto exit ;
				calculateStatisticsForFitnessFunction(leftExtent, rightExtent, i);
				//If the number of nodes prefetched is greater than 20 % of cache size
				if(numberOfNodesPrefetched > this->sizeOfCache*0.10 || count > 5)
					goto exit ;
			} // for (i=levelExtent+delta; (i+delta)<=up; i+=delta)
		} // if (direction==0)		
		if (direction==1)//Indicates prefetch towards the bottom of the tree.
		{
			regionFactory = AscRegionFactory::getInstance();				
			int count = 0;
			for (i=levelExtent-delta; (i-delta)>=down; i-=delta,count++)
			{
				//ger													
		
				QueryModifier qf(1);
				this->loadBrush(leftExtent,rightExtent,i,regionFactory,qf);					
				if ( (i > 1) || (i < 0) )	goto exit ;
				calculateStatisticsForFitnessFunction(leftExtent, rightExtent, i);
				//If the number of nodes prefetched is greater than 20 % of cache size
				if(numberOfNodesPrefetched > this->sizeOfCache*0.10 || count >= 4)
					goto exit ;				
			} // for (i=levelExtent-delta; (i-delta)>=down; i-=delta)
		}
		if (direction==2)//Indicates prefetch towards the left of the tree.
		{
			//Need to make sure that whatever we fetch is in the decreasing the order of x.
			regionFactory = DescRegionFactory::getInstance();
			QueryModifier qf(2);
			this->loadBrush(left,rightExtent,levelExtent,regionFactory,qf);
			calculateStatisticsForFitnessFunction(left, rightExtent, levelExtent);
			goto exit ;
		}
		if (direction==3)//Indicates prefetch towards the right of the tree.
		{  
			//Need to make sure that whatever we fetch is in the increasing order of x.
			regionFactory = AscRegionFactory::getInstance();
			
			QueryModifier qf(1);
			this->loadBrush(leftExtent,right,levelExtent,regionFactory,qf);			
			calculateStatisticsForFitnessFunction(leftExtent, right, levelExtent);
			goto exit ;
		}

		dir[direction] = .0;				

exit:
		::std::cout<<"";

	}catch(PreEmptException pe)
	{
		::std::cout<<"Got preempt exception: "<<pe<<::std::endl;
	}
	catch (...) {
		::std::cout<<"Got exception in do_prefetch : Error"<<::std::endl;
	}				
}

/*
 * @return long, the number of nodes prefetched.
 *
 */
void Prefetcher::loadBrush(double leftExtent,
						   double rightExtent,
						   double lod,
						   AbstractRegionFactory* regionFactory,
						   QueryModifier& queryModifier)
{		
	Brush* brush = new Brush(leftExtent,rightExtent,lod,2,this->brushType);
	this->vecPrefetchBrushes.push_back(brush);			
	LoadRequester *loadRequester;
	PerformanceStatistics perStatistics;
	//LOG_GMSG(prefetchLogger,"Before Calling Load");
	if(leftExtent > rightExtent || lod < 0 || lod > 1) {
		#ifdef _DEBUG
		LOG_GMSG(prefetchLogger,"Error in Prefetch Brush");
		#endif
	}
	try{
		loadRequester = LoadRequester::getInstance();

		this->cacheIterator = 
			loadRequester->load(brush,
			regionFactory,
			perStatistics,
			queryModifier);
	}catch(...)
	{			
		 #ifdef _DEBUG
		LOG_GMSG(prefetchLogger,"Got Exception in Load");
		#endif
	}	
	//If preempt is true then this should end the loop	
	while(!cacheIterator->isDone())
	{	
		//This will lead to calls in the ifContinue()
		cacheIterator->get_next();
		numberOfNodesPrefetched = numberOfNodesPrefetched + 1;
	} // while(!cacheIterator->isDone())		

	//Deleting the cache iterator
	delete(this->cacheIterator);
	this->cacheIterator = NULL;


//	LOG_GMSG(prefetchLogger,"Premept Status "<<preempt);

	//LOG_GMSG(prefetchLogger, "The number of tuples prefetched "<<numberOfNodesPrefetched);
	//Only delete the cacheIterator				

	if(preempt == true)
	{
		#ifdef _DEBUG	
		LOG_GMSG(prefetchLogger,"Get Terminate signal from CacheController");
		#endif
		PreEmptException pe;
		throw(pe);
	}
}

//Set by the Prefetch Controller
void Prefetcher::setSemaphore(ZThread::Semaphore* semPrefetchPermit,
							  ZThread::Semaphore* semLoadPermit)
{
	this->semPrefetchPermit = semPrefetchPermit;
	this->semLoadPermit = semLoadPermit;
}

